#include "boltzman.h"
#include "psolve.h"
#include "ovector.h"
#include "misc.h"
#include "ptclgrp.h"
#include "particle.h"
#include "fields.h"
#include <math.h>

/* Will only work with Multigrid and maybe with DADI - Keith will be happy to fix if you have a problem -
	 send complaints to kc@langmuir.eecs.berkeley.edu
	 */

Boltzmann::Boltzmann(Fields *F, Scalar _temperature, Species* _BoltzSpecies, int _EB_flag)
{
	dump = 0;
	fields = F;
	BoltzSpecies = _BoltzSpecies;
	grid = fields->get_grid();
	psolve = fields->getPoissonSolve();
	ElectrostaticFlag = fields->getElectrostatic();
	bcNode = grid->GetNodeBoundary();

	itermax=100;  //maximum # of iterations for psolve
	tolerance = fields->getpresidue();  //tolerance for the boltzmann solve is set from input file,          

	qMKS = BoltzSpecies->get_q();
	q = qMKS/fabs(ELECTRON_CHARGE);
	J = fields->getJ();
	K = fields->getK();

	temperature = _temperature;
	EB_flag = _EB_flag;  //EB_flag = 1 the energy balance is used 
	                     // otherwise the temperature is constant
	qbytemp = q/temperature;

	foo = new Scalar *[J+1]; 
	phi = new Scalar *[J+1]; 
	rho = new Scalar *[J+1];
	BoltzmannRho = new Scalar *[J+1];
	totalRho = new Scalar *[J+1];
	int j, k;
	for(j=0;j<=J;j++){
		foo[j] = new Scalar[K+1];
		phi[j] = new Scalar[K+1];
		rho[j] = new Scalar[K+1];
		BoltzmannRho[j] = new Scalar[K+1];
		totalRho[j] = new Scalar[K+1];
	}
	for(j=0;j<=J;j++) for(k=0;k<=K;k++){
		phi[j][k] = 0.0;
		foo[j][k] = 0.0;
		BoltzmannRho[j][k] = 0.0;
		totalRho[j][k] = 0.0;
	}

	switch(ElectrostaticFlag) {
	default:
		fprintf(stderr,"WARNING, poisson solve defaulting to Multigrid in Boltzmann Solve\n");

 	case ELECTROMAGNETIC_SOLVE:  
	case DADI_SOLVE:
	case PERIODIC_X1_DADI:
		b1_orig = new Scalar *[J+1];
		b2_orig = new Scalar *[J+1]; 

		for(j=0;j<=J;j++){
			b1_orig[j] = new Scalar[K+1];
			b2_orig[j] = new Scalar[K+1];
		}
		// *1_local, *2_local is shared by psolve and boltzmann
		a1_local = psolve->get_a1_coefficients();
		a2_local = psolve->get_a2_coefficients();
		b1_local = psolve->get_b1_coefficients();
		b2_local = psolve->get_b2_coefficients();
		c1_local = psolve->get_c1_coefficients();
		c2_local = psolve->get_c2_coefficients();
		
		for(j=0;j<=J;j++) for(k=0;k<=K;k++){
			b1_orig[j][k] = b1_local[j][k];
			b2_orig[j][k] = b2_local[j][k];
		}
		break;

		//	case CGXY_SOLVE:
		// morgan
	case CG_SOLVE:
		fprintf(stderr,"WARNING, CG solve has not been written for Boltzmann Solve\n");
		break;
	case GMRES_SOLVE:
	  fprintf(stderr,"WARNING, GMRES solve has not been written for Boltzmann Solve\n");
		break;
	case MGRID_SOLVE:
		resCoeff = new Scalar**[J+1];
		for (j=0;j<=J;j++){
			resCoeff[j] = new Scalar*[K+1];
			for (k=0;k<=K;k++)
				resCoeff[j][k] = new Scalar[5];
		}		
		resCoeff_local = psolve->get_resCoeff();
		for(j=0;j<=J;j++) for(k=0;k<=K;k++)
			for (int s=0; s<5; s++)
				resCoeff[j][k][s] = resCoeff_local[j][k][s];
		break;
	}
}

Boltzmann::~Boltzmann()
{
	for (int j=0; j<=J; j++){
		delete[] b1_orig[j];
		delete[] b1_local[j];
		delete[] phi[j];
		delete[] BoltzmannRho[j];
		delete[] rho[j];
	}
	delete[] b1_orig;
	delete[] b1_local;
	delete[] phi;
	delete[] BoltzmannRho;
	delete[] rho;
}

void Boltzmann::updatePhi(Scalar **source, Scalar **dest, const Scalar t,Scalar dt)
{
	int j,k;
  static int init_flag=1;
	Scalar temp;

	if (init_flag){
		init(source, dest);
		init_flag = 0;
		return;
	}

	for(j=0;j<=J;j++) for(k=0;k<=K;k++)
		phi[j][k] = dest[j][k];
	
	if (totalCharge)
		if (totalCharge*BoltzSpecies->get_q()<0){
			totalCharge = 0;
			energy = 0;
		}

	if (totalCharge)
		if (temperature<.05){
			totalCharge = 0.0;
			temperature = 6;
			energy = 0;
		}
	
	if (totalCharge){
		if (NonlinearSolve(source)){
			BoltzSpecies->zeroKineticEnergy();
			BoltzSpecies->addKineticEnergy(3.0/2.0*temperature*totalCharge/q);


			Vector2 deltaC = Vector2(0,0); 
//			deltaC = Collisions(dt);

			// particle and energy flux to wall
			deltaC += dt*ParticleFlux(dt);
			
#ifdef BOLT_DEBUG
			cout << "Particle Flux " << deltaC.e1()/qMKS/1e7 << "; Particles " << totalCharge/qMKS/1e7 <<  " Fractonal change " << deltaC.e1()/totalCharge << endl;
			cout << "Kinetic Energy Flux " << 5.0/2.0*temperature*deltaC.e1()/qMKS/1e7 << "; Energy " << energy/1e7 << endl;
			cout << "Pot Energy Flux " << deltaC.e2()/1e7 << endl;
#endif

			totalCharge += deltaC.e1();
			energy += 5.0/2.0*temperature*deltaC.e1()/qMKS;
			energy += deltaC.e2();
			
		}
		else{
			LinearSolve(source);
			BoltzSpecies->zeroKineticEnergy();
			BoltzSpecies->addKineticEnergy((3.0/2.0*totalCharge*temperature)/q);
		}
	}
	else{
		LinearSolve(source);
		BoltzSpecies->zeroKineticEnergy();
//		BoltzSpecies->addKineticEnergy((3.0/2.0*totalCharge*temperature)/q);
	}
	for(j=0;j<=J;j++) for(k=0;k<=K;k++)
		dest[j][k]=phi[j][k];
	
}

void Boltzmann::init(Scalar **source, Scalar **dest)
{
	int j,k;
	Scalar PotEnergy = 0;
	Scalar totalVolume = 0;
	int EBFlag_orig;

	for(j=0;j<=J;j++) for(k=0;k<=K;k++)
		if (!phi[j][k])
			phi[j][k] = dest[j][k];

	if (!dump){
#ifdef NON_N
		totalCharge = 0.0;
#else
		// set the init totalCharge equal - particle density
		totalCharge=0.0; 
		for(j=0;j<=J;j++) 
			for(k=0;k<=K;k++) 
				totalCharge+=source[j][k]*grid->get_halfCellVolumes()[j][k];
		totalCharge *= q; 
#endif
		LinearSolve(source);
		
		energy = temperature*3.0/2.0*totalCharge; 
		EBFlag_orig = EB_flag;
		EB_flag = 0;
		// Start off with this temperture and find the PotEnergy
		bRho = fields->getRhoSpecies()[BoltzSpecies->getID()];
		
		if (totalCharge)
			NonlinearSolve(source);
		else
			LinearSolve(source);
		EB_flag = EBFlag_orig;
	}
	else{
		bRho = fields->getRhoSpecies()[BoltzSpecies->getID()];
		
		if (totalCharge)
			NonlinearSolve(source);
		else
			LinearSolve(source);
	}

	for(j=0;j<=J;j++) for(k=0;k<=K;k++)
		dest[j][k]=phi[j][k];

}

void Boltzmann::PSolveCoeff(Scalar **source)
{
	
	Scalar temp;
	int j,k;

	switch(ElectrostaticFlag) {
	default:
		fprintf(stderr,"WARNING, poisson solve defaulting to Multigrid in Boltzmann Solve\n");
		
 	case ELECTROMAGNETIC_SOLVE:  
	case DADI_SOLVE:
	case PERIODIC_X1_DADI:
		for(j=0;j<=J;j++) for(k=0;k<=K;k++){
			temp = -BoltzmannRho[j][k];
			//		temp=-n0*exp(-phi[j][k]*qbytemp);
			rho[j][k] = -source[j][k]+(1+qbytemp*phi[j][k])*temp;
			if (b1_orig[j][k])
				b1_local[j][k] = b1_orig[j][k]+0.5*qbytemp*temp;
			if (b2_orig[j][k])
				b2_local[j][k] = b2_orig[j][k]+0.5*qbytemp*temp;
		}
		break;
	case CG_SOLVE:
	case GMRES_SOLVE:
		fprintf(stderr,"WARNING, CG and GMRES solves has not been written for Boltzmann Solve\n");
		break;
	case MGRID_SOLVE:
		psolve->PSolveBoltzCoeff(n0, qbytemp, BoltzmannRho, MinPot);
		for(j=0;j<=J;j++) for(k=0;k<=K;k++){
			temp = -BoltzmannRho[j][k];
			rho[j][k] = -source[j][k]+(1+qbytemp*phi[j][k])*temp;
		}
		break;
	}
}

Scalar Boltzmann::Residual()
{
	Scalar res = 0.0;
	Scalar temp = 0.0;
	
	switch(ElectrostaticFlag) {
	default:
		fprintf(stderr,"WARNING, poisson solve defaulting to Multigrid in Boltzmann Solve\n");
		
 	case ELECTROMAGNETIC_SOLVE:  
	case DADI_SOLVE:
		int j,k;
		
		j=0;k=0;
		if (b1_orig[j][k] !=0)  {
			temp += totalRho[j][k];
			res += sqr(b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
								 b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]+
								 totalRho[j][k]);
			foo[j][k] = sqr(b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
											b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]+
											totalRho[j][k]); 
		}

		k=K;
		if (b1_orig[j][k] !=0){
			temp += totalRho[j][k];
			res += sqr(b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
								 a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+
								 totalRho[j][k]);
			foo[j][k] = sqr(b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
											a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+
											totalRho[j][k]);  
		}

		j=J;
		if (b1_orig[j][k] !=0){
			temp += totalRho[j][k];
			res += sqr(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+
								 a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+
								 totalRho[j][k]);
			foo[j][k] = sqr(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+
											a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+
											totalRho[j][k]); 
		}
		
		k=0;

		if (b1_orig[j][k] !=0){
			temp += totalRho[j][k];
			res += sqr(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+
								 b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]+
								 totalRho[j][k]);
			foo[j][k] = sqr(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+
											b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]+
											totalRho[j][k]);
		}

		for (j=1;j<J;j++){
			k = 0;
			if (b1_orig[j][k] !=0){
				temp += totalRho[j][k];
				res += sqr(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
									 b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]+
									 totalRho[j][k]);
				foo[j][k] = sqr(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
												b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]+
												totalRho[j][k]);
			}
			k = K;
			if (b1_orig[j][k] !=0){
				temp += totalRho[j][k];
				res += sqr(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
									 a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+
									 totalRho[j][k]);
				foo[j][k] = sqr(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
												a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+
												totalRho[j][k]);
			}
		}

		for(k=1;k<K;k++){
			j = 0;
			if (b1_orig[j][k] !=0){
				temp += totalRho[j][k];
				res += sqr(b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
									 a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]+
									 totalRho[j][k]);
				foo[j][k] = sqr(b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
												a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]+
												totalRho[j][k]);
			}
			j = J;
			if (b1_orig[j][k] !=0){
				temp += totalRho[j][k];
				res += sqr(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+
									 a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]+
									 totalRho[j][k]);
				foo[j][k] = sqr(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+
												a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]+
												totalRho[j][k]);
			}
		}

		for (j=1;j<J;j++) for(k=1;k<K;k++){		
			res += sqr(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
								 a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]+
								 totalRho[j][k]);
			foo[j][k] = sqr(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
											a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]+
											totalRho[j][k]);
			
			/*		foo[j][k] = -(a1_local[j][k]*phi[j-1][k]+b1_orig[j][k]*phi[j][k]+c1_local[j][k]*phi[j+1][k]+
						a2_local[j][k]*phi[j][k-1]+b2_orig[j][k]*phi[j][k]+c2_local[j][k]*phi[j][k+1]);*/
			if (b1_orig[j][k]!=0)
				temp += sqr(totalRho[j][k]);
		}
		if (temp){
			res = sqrt(res/temp);
			for (j=0;j<=J;j++) for(k=0;k<=K;k++)
				foo[j][k] = sqrt(foo[j][k]/temp);
			
		}
		
		break;
	case CG_SOLVE:
	case GMRES_SOLVE:
		fprintf(stderr,"WARNING, CG and GMRES solves has not been written for Boltzmann Solve\n");
		break;
	case MGRID_SOLVE:
		for (j=0;j<=J;j++)
			for (k=0;k<=K;k++)
				resCoeff_local[j][k][4] = resCoeff[j][k][4];
		
		res = psolve->Resid(totalRho, phi);	
	}	
	return (res);
}
Vector2 Boltzmann::ParticleFlux(Scalar dt)
{
	Scalar PartCurrent;
	Scalar EnergyCurrent;
	Scalar current;
	int j,k;
	Scalar foofoo;

	Boundary* bPtr;
	Boundary*** bc1 = grid->GetBC1();
	Boundary*** bc2 = grid->GetBC2();
	Vector3** dS = grid->dS();

	const Scalar scale = dt/BoltzSpecies->get_q()*.5*sqrt(2*temperature*fabs(ELECTRON_CHARGE)/M_PI/BoltzSpecies->get_m());

	PartCurrent = 0;
	EnergyCurrent = 0;

	for (j=1;j<=J;j++) 
		for(k=1;k<=K;k++){
			if ((bPtr=bcNode[j][k])){
				if ((bPtr=bc1[j-1][k]))
					current = Collect(bPtr, .5*dS[j-1][k].e2(), Vector2(j,k),  BoltzmannRho[j][k], scale);
				if ((bPtr=bc2[j][k-1]))
					current += Collect(bPtr, .5*dS[j][k-1].e1(),  Vector2(j,k), BoltzmannRho[j][k], scale);
				PartCurrent += current;
				EnergyCurrent += phi[j][k]*current;
			}
		}

	for (j=1;j<J;j++) 
		for(k=1;k<K;k++){
			if ((bPtr=bcNode[j][k])){
				if ((bPtr=bc1[j][k]))
					current = Collect(bPtr,.5*dS[j][k].e2(), Vector2(j,k), BoltzmannRho[j][k],scale);
				if ((bPtr=bc2[j][k]))
					current += Collect(bPtr, .5*dS[j][k].e1(),  Vector2(j,k), BoltzmannRho[j][k], scale);
				PartCurrent += current;
				EnergyCurrent += phi[j][k]*current;
			}
		}
	
	j=0; k=0;
	if ((bPtr=bcNode[j][k])){
		if ((bPtr=bc1[j][k]))
			current = Collect(bPtr, .5*dS[j][k].e2(),  Vector2(j,k), BoltzmannRho[j][k], scale);
		if ((bPtr=bc2[j][k]))
			current += Collect(bPtr, .5*dS[j][k].e1(),  Vector2(j,k), BoltzmannRho[j][k], scale);
		PartCurrent += current;
		EnergyCurrent += phi[j][k]*current;
	}


	j=0;
	for(k=1;k<K;k++){
		if ((bPtr=bcNode[j][k])){
			if ((bPtr=bc1[j][k]))
				current = Collect(bPtr, .5*dS[j][k].e2(),  Vector2(j,k), BoltzmannRho[j][k], scale);
			if ((bPtr=bc2[j][k]))
				current += Collect(bPtr, .5*dS[j][k].e1(),  Vector2(j,k), BoltzmannRho[j][k], scale);
			if ((bPtr=bc2[j][k-1]))
				current += Collect(bPtr, .5*dS[j][k-1].e1(),  Vector2(j,k), BoltzmannRho[j][k], scale);
			PartCurrent += current;
			EnergyCurrent += phi[j][k]*current;
		}
	}
	
	if ((bPtr=bcNode[j][k]))
		if ((bPtr=bc2[j][k-1])){
			current = Collect(bPtr, .5*dS[j][k-1].e1(),  Vector2(j,k), BoltzmannRho[j][k], scale);
			PartCurrent += current;
			EnergyCurrent += phi[j][k]*current;
		}
	
	k=0;
	for (j=1;j<J;j++){ 
		if ((bPtr=bcNode[j][k])){
			if ((bPtr=bc1[j][k]))
				current = Collect(bPtr, .5*dS[j][k].e2(),  Vector2(j,k), BoltzmannRho[j][k], scale);
			if ((bPtr=bc1[j-1][k]))
				current = Collect(bPtr, .5*dS[j-1][k].e2(),  Vector2(j,k), BoltzmannRho[j][k], scale);
			if ((bPtr=bc2[j][k]))
				current = Collect(bPtr, .5*dS[j][k].e1(),  Vector2(j,k), BoltzmannRho[j][k], scale);
			PartCurrent += current;
			EnergyCurrent += phi[j][k]*current;
		}
	}

	if ((bPtr=bcNode[j][k]))
		if ((bPtr=bc1[j-1][k])){
			current = Collect(bPtr, .5*dS[j-1][k].e2(),  Vector2(j,k), BoltzmannRho[j][k], scale);
			PartCurrent += current;
			EnergyCurrent += phi[j][k]*current;
		}
	
	PartCurrent *= .5*sqrt(2*temperature*fabs(ELECTRON_CHARGE)/M_PI/BoltzSpecies->get_m());  // don't use q_over_m because fabs(ELECTRON_CHARGE) is the converstion from eV to J
	EnergyCurrent *= .5*sqrt(2*temperature/fabs(ELECTRON_CHARGE)/M_PI/BoltzSpecies->get_m());  // don't use q_over_m because fabs(ELECTRON_CHARGE) is the converstion from eV to J

	return (Vector2(PartCurrent, EnergyCurrent)); 
	
}

Scalar Boltzmann::Collect(Boundary* bPtr, Scalar area, const Vector2& x,const Scalar& edge_rho, const Scalar& scale)
{
	Scalar Flux=0;

	Flux = area*edge_rho*bPtr->getBoltzmannFlux();
	if (Flux){
		Scalar periodicFactor=1;
		if (grid->getPeriodicFlagX1()&&(x.e1()==0||x.e1()==J))
			periodicFactor = .5;
		if (grid->getPeriodicFlagX2()&&(x.e2()==0||x.e2()==K))
			periodicFactor = .5;

				
		bPtr->collect(*new Particle(x, Vector3(0,0,0), BoltzSpecies, fabs(periodicFactor*Flux*scale),1),
									*new Vector3(0,0,0));  // varweight particle  
	}

	return (Flux);
}

ParticleList& Boltzmann::testParticleList(ParticleGroupList& pgList)
{
	//	ParticleList pList;
	if (!pgList.head) return pList;
	Species* species = pgList.head->data->get_species();
	ParticleGroup* pg;
	Scalar phi0;

	getMinPot();

	int j,k;
	
	oopicListIter<ParticleGroup> pgIter(pgList);
	for (pgIter.restart(); !pgIter.Done(); pgIter++){
		pg = pgIter();
		for (int index=0; index<pgIter()->get_n(); index++){
			//bilinear interpolation
			phi0 = grid->interpolateBilinear(phi, pg->get_x(index));
			
			if ((MinPot-phi0+pg->energy_eV(index))<species->get_threshold()){
				pList.add(new Particle(pg->get_x(index), pg->get_u(index), 
															 fields->get_bSpecies(), 
															 pg->get_np2c(index), 1));  // make the Boltzmann pg varweight.
				
				//	Move last particle into this slot
				pg->fill(index);
				index--;
			}
		}
	}
	return pList;
}

void Boltzmann::IncBoltzParticles(ParticleGroupList& pgList)
{

	if (!pgList.head) return;

	Scalar charge = 0;
	Scalar KEint = 0;
	Scalar phi0=0;
	Scalar Denergy = 0;
	Scalar KE = 0;
	ParticleGroup* pg;
	int j,k; //,index;
	int i = BoltzSpecies->getID();
	
	getMinPot();

	oopicListIter<ParticleGroup> pgIter(pgList);
	for (pgIter.restart(); !pgIter.Done(); pgIter++){ 
		pg = pgIter();
		for (int index=(pgIter()->get_n()-1); index>=0; index--){
			phi0 = grid->interpolateBilinear(phi, pg->get_x(index));
			charge = pg->get_q(index);
			//			Denergy = charge*(MinPot-phi0+pg->energy_eV(index)); 
			KE = pg->energy_eV(index);
			Denergy = charge*(-phi0+pg->energy_eV(index))/qMKS; 
			Scalar FE = charge*(MinPot-phi0+pg->energy_eV(index))/qMKS; 
			energy += Denergy;
			totalCharge += charge;
#ifdef BOLT_DEBUG			
			cout << "KE "<< KE << " FE " << FE/1e7 << " Denergy " << Denergy/1e7 << "; Charge " << charge/qMKS/1e7 << endl; 
#endif
		}
		pg->set_n(0);
	}

}

void Boltzmann::getMinPot(void)
{
	// if sign == 1 return a min
	// if sign == -1 return a max
	// need to add something here to see if phi has changed
	//	Scalar MinPot=sign*10000;
	MinPot = q*10000;
	int sign = (int)(q/fabs(q));

	for(int j=0;j<=J;j++) for(int k=0;k<=K;k++)
		if (grid->PlasmaRegion(j,k))
			if ((sign*MinPot) > (sign*phi[j][k]))
				MinPot = phi[j][k];
	
}

int Boltzmann::Dump(FILE* DMPFile)
{

	XGWrite(&totalCharge, ScalarInt, 1, DMPFile, ScalarChar);
	XGWrite(&energy, ScalarInt, 1, DMPFile, ScalarChar);
	for (int j=0; j<=J;j++)
		for (int k=0; k<=K;k++)
			XGWrite(&phi[j][k], ScalarInt, 1, DMPFile, ScalarChar);

	return (1);
}

int Boltzmann::Restore(FILE* DMPFile)
{
	
	dump = 1;

	XGRead(&totalCharge, ScalarInt, 1, DMPFile, ScalarChar);
	XGRead(&energy, ScalarInt, 1, DMPFile, ScalarChar);
	for (int j=0; j<=J;j++)
		for (int k=0; k<=K;k++)
			XGRead(&phi[j][k], ScalarInt, 1, DMPFile, ScalarChar);

	return (1);
}

int Boltzmann::NonlinearSolve(Scalar** source)
{
	int j,k;
	Scalar res=10;
	int it=0;
	int it_out = 0;
	int itmax=20;
	Scalar temp;
	int EB;
	Scalar pstol = .1;  //tolerance for the poisson solver => tolerance*pstol

	getMinPot();  //Finds the min pot within the plasma
	UpdateBRho();

	for(j=0;j<=J;j++) for(k=0;k<=K;k++)
		totalRho[j][k] = source[j][k] + BoltzmannRho[j][k];
	
	EBCFlag = 0;
	PSCFlag = 0;

	while ((EBCFlag==0)&&(PSCFlag==0)&&(it_out < (itmax/5))){
		while ((EBCFlag==0)&&(PSCFlag==0)&&(it < itmax)){
			
			PSolveCoeff(source); // change psolve coefficients
			psolve->solve(phi,rho,itermax,pstol*tolerance); 
			
//			for(j=0;j<=J;j++) for(k=0;k<=K;k++)
//				phi[j][k] = -150 + j*300/J;

//			cout << "UpdateBRho " << endl;
			getMinPot();
			UpdateBRho();

			for(j=0;j<=J;j++) for(k=0;k<=K;k++)
				totalRho[j][k] = source[j][k] + BoltzmannRho[j][k];
			res = Residual();

			if (res < tolerance)
				PSCFlag = 1;

#ifdef BOLT_DEBUG		
			cout << it << " " << res << endl;
#endif
				
			EB = EnergyBalance();
			if (EB==1){
				for(j=0;j<=J;j++) for(k=0;k<=K;k++)
					totalRho[j][k] = source[j][k] + BoltzmannRho[j][k];
				res = Residual();
				it++;
			
#ifdef BOLT_DEBUG			
				cout << it << " " << res << endl;
				cout << " Kinetic Energy " << totalCharge*3.0*temperature/(2.0*qMKS*1e7) << 
					" Total Energy " << energy/1e7 << endl;
#endif
			}
			else if (EB==2)
				return (0);
		}
		if (res > tolerance){
			cout << "Nonlinear Boltzmann Solve FAILURE: Residual = " << res << endl;
			pstol *=.5;
			it_out++;
			cout << "Decreasing psolve tolerance to " << pstol*tolerance << endl;
			it = 0;
		}
	}
	if (res > tolerance)
		cout << "Nonlinear Boltzmann Solve TOTAL FAILURE: Residual = " << res << endl;
	
#ifdef BOLT_DEBUG
	cout << "Nonlinear Solved" << endl;
	cout << " Residual " << res << " Kinetic Energy " << totalCharge*3.0*temperature/(2.0*qMKS*1e7) << 
					" Total Energy " << energy/1e7 <<  endl;
#endif			

	return (1);
}

void Boltzmann::LinearSolve(Scalar** source)
{
	int j,k;
	
	for(j=0;j<=J;j++) for(k=0;k<=K;k++) 
		rho[j][k] = -source[j][k];
	
	switch(ElectrostaticFlag) {
	default:
		fprintf(stderr,"WARNING, poisson solve defaulting to Multigrid in Boltzmann Solve\n");
		
	case ELECTROMAGNETIC_SOLVE:  
	case DADI_SOLVE:
	case PERIODIC_X1_DADI:
		for(j=0;j<=J;j++) for(k=0;k<=K;k++){
			if (b1_orig[j][k])
				b1_local[j][k] = b1_orig[j][k];
			if (b2_orig[j][k])
				b2_local[j][k] = b2_orig[j][k];
		}
		break;
	case CG_SOLVE:
	case GMRES_SOLVE:
		fprintf(stderr,"WARNING, CG and GMRES solves has not been written for Boltzmann Solve\n");
		break;
	case MGRID_SOLVE:
		psolve->PSolveBoltzCoeff(0, qbytemp, phi, 0);
	}

	psolve->solve(phi,rho,itermax,tolerance);

	for(j=0;j<=J;j++) for(k=0;k<=K;k++){
		BoltzmannRho[j][k] = 0.0;
		totalRho[j][k] = source[j][k] + BoltzmannRho[j][k];
	}

}

void Boltzmann::UpdateBRho()
{
	int j,k;
	Scalar sum_of_exp = 0.0;

	for(j=0;j<=J;j++) 
		for(k=0;k<=K;k++){
			if (grid->PlasmaRegion(j,k)){
				if ((-(phi[j][k]-MinPot)*qbytemp)<-50.0)
					BoltzmannRho[j][k] = 0;
				else
					BoltzmannRho[j][k] = exp(-(phi[j][k]-MinPot)*qbytemp);
				sum_of_exp += (BoltzmannRho[j][k]*grid->get_halfCellVolumes()[j][k]);
			}
			else 
				BoltzmannRho[j][k] = 0.0;
		}

	n0 = totalCharge/sum_of_exp;
	
	for(j=0;j<=J;j++) for(k=0;k<=K;k++)
		BoltzmannRho[j][k] *= n0;
	
	return;
}

int Boltzmann::EnergyBalance()  // returns one if balanced
{
	int j,k;
	Scalar PotEnergy = 0.0;
	Scalar PotEnergyDiv = 0.0;
	Scalar PEgrid = 0.0;
	Scalar deltaT = 0.0;

	Scalar OldTemp = temperature;
	
	if (EB_flag){
		for(j=0;j<=J;j++) for(k=0;k<=K;k++){ 
			PEgrid =  phi[j][k]*BoltzmannRho[j][k]*grid->get_halfCellVolumes()[j][k];
			PotEnergy += PEgrid;
			PotEnergyDiv += PEgrid*(phi[j][k]-MinPot);
		}
		PotEnergy /= fabs(ELECTRON_CHARGE); // J->eV
		PotEnergyDiv *= qbytemp/temperature;
		PotEnergyDiv /= fabs(ELECTRON_CHARGE); // J->eV

		if (PSCFlag&&(fabs((totalCharge*3.0*temperature/qMKS/2.0+PotEnergy - energy)/energy)<tolerance)){
			EBCFlag = 1;
		}
		else{
			PSCFlag = 0;
			Scalar NewTemp = 2.0*(energy - PotEnergy)*qMKS/totalCharge/3.0;
			temperature = (NewTemp+temperature)/2;
//			temperature = 2.0*(energy - PotEnergy)*qMKS/totalCharge/3.0;
/*			deltaT= (PotEnergy-energy+3*totalCharge/(2*qMKS)*temperature)/
				(3*totalCharge/(2*qMKS) + PotEnergyDiv);
			while ((deltaT>0)&&(deltaT>temperature)) deltaT/=2;
			temperature -= deltaT;
*/																																				
				

#ifdef BOLT_DEBUG
			cout << "Pot Energy " << PotEnergy/1e7 << 
				" Kinetic Energy " << totalCharge*3.0*temperature/(2.0*qMKS*1e7) << 
					" Total Energy " << energy/1e7 << endl;
			cout << "Energy Residual " << fabs((3.0*totalCharge*temperature/qMKS/2.0+PotEnergy - energy)/energy) << endl;
#endif
			
			if (temperature<=0){  
				// if the temperature goes negative there is not
				// enought thermal energy for this method to converge
				cout << "temperature went negative.  " << temperature << endl;
//				temperature = OldTemp/2;
//				cout << "Dividing old temperature by 2.  " << OldTemp << endl;
//				qbytemp = q/temperature;
				qbytemp = -1;
				totalCharge =0;
				energy = 0;
				temperature = 0;
				return 2;
//				return 0;
			}
			else{
				qbytemp = q/temperature;
				EBCFlag = 0;
				return 0;
			}
		}
	}
	else
		EBCFlag = 1;
	
	return 1;
}

Vector2 Collisions(Scalar dt)
{
	
	
	return Vector2(0,0);
}
