/*====================================================================

GPDIST.CPP

  A class which accepts a particle and maintains a space and
velocity distribution.  Meant to be used in boundaries with collect.

0.99	(PeterM ?)  Inception
0.991 (PeterM 8-29-95)  Add a method for computing Q, the sum of the
      charge which has hit the wall, + misc other changes.
1.01  (JohnV 05-16-97) Add ftheta.
2.0   (JohnV 09-30-97) Revamped to store only f(x); should be fixed later.
2.1   (JohnV 06-28-98) Fixing later to include f(x,energy) and f(x,theta).

====================================================================
*/

#include "ovector.h"
#include "particle.h"
#include "gpdist.h"
#include "grid.h"
#include "ostring.h"

PDistCol::PDistCol(int _nxbins,	 int _nenergybins, int _nthetabins, int _nsp,
						 Scalar _energy_min, Scalar _energy_max,
						 Scalar _x_min,  Scalar _x_max,
						 Scalar _theta_min, Scalar _theta_max/*, Species *diag_species*/)
{ 
	nxbins = _nxbins;
	nxpoints = nxbins + 1;
	x_min = _x_min;
	x_max = _x_max;
	nenergybins = _nenergybins;
	nepoints = ntpoints = 0;
	if (nenergybins) nepoints = nenergybins + 1;
	energy_min = _energy_min;
	energy_max = _energy_max;
	nthetabins = _nthetabins;
	if (nthetabins) ntpoints = nthetabins + 1;
	theta_min = _theta_min*M_PI/180; // convert to radians internally
	theta_max = _theta_max*M_PI/180;
	nsp = _nsp;
   int i, j;
	fxdata = new Scalar *[nsp];
	x = new Scalar[nxbins+1];
	Qlast = Qcol = 0;

	if(!nxbins) { nsp=0;return;};
	// we don't need to allocate diagnostics if nxbins==0

	for(i=0; i<nsp; i++)	{  
	  fxdata[i] = new Scalar[nxbins+1];
	  memset(fxdata[i], 0, (nxbins+1)*sizeof(Scalar));
	}
	xscale = nxbins/(x_max - x_min);
	for (i=0; i <= nxbins; i++) x[i]= i/xscale + x_min;

	if (nenergybins > 0){
	  fedata = new Scalar **[nsp];
	  for (i=0; i<nsp; i++){
		 fedata[i] = new Scalar *[nxbins+1];
		 for (j=0; j<=nxbins; j++){
			fedata[i][j] = new Scalar[nenergybins+1];
			memset(fedata[i][j], 0, (nenergybins+1)*sizeof(Scalar));
		 }
	  }
	  escale = nenergybins/(energy_max - energy_min);
	  energy = new Scalar[nenergybins+1];
	  for (j=0; j<=nenergybins; j++) energy[j] = energy_min + j/escale;
	}

	if (nthetabins > 0){
	  ftdata = new Scalar **[nsp];
	  for (i=0; i<nsp; i++){
		 ftdata[i] = new Scalar *[nxbins+1];
		 for (j=0; j<=nxbins; j++){
			ftdata[i][j] = new Scalar[nthetabins+1];
			memset(ftdata[i][j], 0, (nthetabins+1)*sizeof(Scalar));
		 }
	  }
	  tscale = nthetabins/(theta_max - theta_min);
	  theta = new Scalar [nthetabins+1];
	  for (j=0; j<=nthetabins; j++) theta[j] = theta_min + j/tscale;
	}
}

/* component is which of the two positions to use for collection */
void PDistCol::record_particle(Particle &p, int component)
{
   //collect the charge of this particle
	Qcol+=p.get_q();

   if(!nxbins) return;
   int spec = p.get_speciesID();
	Scalar xp = (component)? p.get_x().e2(): p.get_x().e1();
	xp = (xp-x_min)*xscale;
	if (xp > nxbins+1 || xp < 0) return;
	
	/* actually do the accumulation! */
	/* this assumes that the particle has the position on the boundary given by xp*/
	int j = (int) xp; // the floor index
	Scalar w = xp - j; // the fraction within the bin
	fxdata[spec][j] += p.get_np2c()*(1-w);
	fxdata[spec][j+1] += p.get_np2c()*w;
	
	if (nenergybins){
	  Scalar e = (p.energy_eV_MKS() - energy_min)*escale;
	  if (e > 0 && e < nenergybins) {
		 int k = (int) e;
		 Scalar we = e - k; // fraction in bin
		 if (w){
			 fedata[spec][j][k] += p.get_np2c()*(1-w)*(1-we);
			 fedata[spec][j][k+1] += p.get_np2c()*(1-w)*we;
			 fedata[spec][j+1][k] += p.get_np2c()*w*(1-we);
			 fedata[spec][j+1][k+1] += p.get_np2c()*w*we;
		 }
		 else{
			 fedata[spec][j][k] += p.get_np2c()*(1-we);
			 fedata[spec][j][k+1] += p.get_np2c()*we;
		 }
	 }
	}
	
	if (nthetabins){
	  Scalar t;
	  Vector3 u = p.get_u();
	  if (component) t = atan(sqrt(u.e2()*u.e2() + u.e3()*u.e3())/u.e1());
	  else t = atan(sqrt(u.e1()*u.e1() + u.e3()*u.e3())/u.e2());
	  t = (t - theta_min)*tscale;
	  if (t > 0 && t < nthetabins){
		 int k = (int) t;
		 Scalar wt = t - k;
		 if (w){
			 ftdata[spec][j][k] += p.get_np2c()*(1-w)*(1-wt);
			 ftdata[spec][j][k+1] += p.get_np2c()*(1-w)*wt;
			 ftdata[spec][j+1][k] += p.get_np2c()*w*(1-wt);
			 ftdata[spec][j+1][k+1] += p.get_np2c()*w*wt;
		 }
		 else{
			 ftdata[spec][j][k] += p.get_np2c()*(1-wt);
			 ftdata[spec][j][k+1] += p.get_np2c()*wt;
		 }
	 }
	}
}

PDistCol::~PDistCol() {
  delete[] x;
  for (int i=0; i<nsp; i++){
	 delete[] fxdata[i];
	 for (int j=0; j<=nxbins; j++){
		if (nenergybins) delete[] fedata[i][j];
		if (nthetabins) delete[] ftdata[i][j];
	 }
	 if (nenergybins) delete[] fedata[i];
	 if (nthetabins) delete[] ftdata[i];
  }
  delete fxdata;
  delete fedata;
  delete ftdata;
}

#ifdef oldcode
============================================================================
PDistCol::PDistCol(int _nxbins,	 int _nvbins, int _nthetabins, int _nsp,
						 Vector3 _vstart, Vector3 _vend,
						 Scalar _xstart,  Scalar _xend,
						 Scalar _theta_min, Scalar _theta_max,
						 Species *diag_species)
{ 
	int i,j;

	xend = _xend;
	xstart = _xstart;
	theta_min = _theta_min*M_PI/180; // use rad internally
	theta_max = _theta_max*M_PI/180;
	nxbins = _nxbins;
	nvbins = _nvbins;
	nthetabins = _nthetabins;
	nsp = _nsp;
	data = new Vector3 **[nsp];
	vstart = new Vector3 [nsp];
	vend = new Vector3 [nsp];
	vscale = new Vector3 [nsp];
	spf = new Species * [nsp];
	x = new Scalar [nxbins];
	v = new Vector3 *[nsp];
	if (nthetabins > 0) {
	  theta = new Scalar[nthetabins+1];
	  ftheta = new Scalar **[nsp];
	}

	Qlast = Qcol = 0;

	if(!nxbins) { nsp=0;return;};
	// we don't need to allocate diagnostics if nxbins==0

   int spec=diag_species->getID();
	vstart[spec]=_vstart;
	vend[spec]=_vend;
	dsp = spec;

	for(i=0;i<nsp;i++)
		{  spf[i] = 0;
			data[i]= new Vector3 *[nxbins];
			if (nthetabins > 0) ftheta[i] = new Scalar *[nxbins];
			v[i] = new Vector3[nvbins];
			for(j=0;j<nxbins;j++) {
				data[i][j] = new Vector3[nvbins];
				/* zero the arrays */
				memset(data[i][j],0,nvbins*sizeof(Vector3));
				if (nthetabins > 0) {
				  ftheta[i][j] = new Scalar[nthetabins];
				  memset(ftheta[i][j], 0, (nthetabins+1)*sizeof(Scalar));
				}
			}
			memset(v[i],0,nvbins*sizeof(Vector3));
		}

	vscale[spec] = Vector3(1.0,1.0,1.0).jvDivide((1.0/(Scalar)nvbins) * (vend[spec] - vstart[spec]));
	xscale = nxbins/(xend-xstart);
	spf[spec] = diag_species;
	for( i=0;i<nxbins;i++) x[i]= i/xscale + xstart;

  /*  set up the velocity array for species spec, the others must be done later */
	for(i=0;i<nvbins;i++) 
		v[spec][i] = Vector3((Scalar) i,(Scalar) i,(Scalar) i).jvDivide(vscale[spec]) +
			vstart[spec];

	if (nthetabins > 0) {
	  thetaScale = nthetabins/(theta_max - theta_min);
	  for (i=0; i<=nthetabins; i++) theta[i] = i/thetaScale + theta_min;
	}
}


/* component is which of the two positions to use for collection */
void PDistCol::record_particle(Particle &p, int component)
{
	int spec=p.get_speciesID();
	Scalar x = (component)? p.get_x().e2(): p.get_x().e1();
	int xi;  /*  the position in the space array */
	Vector3 vi;
	Vector3 *tmp;

   //collect the charge of this particle
	Qcol+=p.get_q();

   if(!nxbins) return;
	/* uninitialized vstart, vend, so we set them, assuming like temperatures. */
	if(!spf[spec]) 
		{ 
			spf[spec]=p.get_speciesPtr();
			Scalar cal_mass = spf[dsp]->get_m();
			vstart[spec]=vstart[dsp] * sqrt(cal_mass/spf[spec]->get_m());
			vend[spec]=vend[dsp] * sqrt(cal_mass/spf[spec]->get_m());
			vscale[spec] = Vector3(1.0,1.0,1.0).jvDivide((1.0/(Scalar)nvbins) 
												* (vend[spec] - vstart[spec]));
			for(int i=0;i<nvbins;i++) 
				v[spec][i] = Vector3((Scalar) i,(Scalar) i,(Scalar) i).jvDivide(
											vscale[spec]) +vstart[spec];

		}

	
	/* actually do the accumulation! */
	/* this assumes that the particle has the position on the boundary given by p*/
	xi = (int) ((x - xstart)*xscale);
	if(xi>=nxbins||xi<0) return;
	vi = p.get_u();

	// if nthetabins > 0 compute angular dist; currently ignore v3
	if (nthetabins > 0) {
	  Scalar theta;
	  if (component) theta = atan(vi.e2()/(vi.e1())+1e-20); // component == alongx2()
	  else theta = atan(vi.e1()/(vi.e2()+1e-20));
	  if (theta >= theta_min && theta <= theta_max)
		 ftheta[spec][xi][(int) (0.5+thetaScale*(theta - theta_min))] += p.get_np2c();
	}

	vi -= vstart[spec];
	vi = vi.jvMult(vscale[spec]) + Vector3(0.5,0.5,0.5);
	//if(vi.e1()>=nvbins||vi.e2()>=nvbins||vi.e3()>=nvbins||vi.e1()<0||vi.e2()<0||vi.e3()<0) return;

	/* store the recorded v */
	if(vi.e1()<nvbins && vi.e1() >= 0) {
		tmp = &data[spec][xi][(int)vi.e1()];
		tmp->set_e1(tmp->e1()+p.get_np2c());
	}
	if(vi.e2()<nvbins && vi.e2() >= 0) {
		tmp = &data[spec][xi][(int)vi.e2()];
		tmp->set_e2(tmp->e2()+p.get_np2c());
	}

	if(vi.e3()<nvbins && vi.e3() >= 0) {
		tmp = &data[spec][xi][(int)vi.e3()];
		tmp->set_e3(tmp->e3()+p.get_np2c());
	}
}

PDistCol::~PDistCol(void) {
	delete[] vstart;
	delete[] vend;
	delete[] vscale;
	delete[] spf;
	delete[] x;
	if (nthetabins > 0) delete[] theta;
	for(int i=0;i<nsp;i++)
		{
			delete[] v[i];
			for(int j=0;j<nxbins;j++) {
			  delete[] data[i][j];
			  if (nthetabins > 0) delete[] ftheta[i][j];
			}

			delete[] data[i];
			if (nthetabins > 0) delete[] ftheta[i];
		}
}			

#endif 
