/*====================================================================

GPDIST.H

  A class which accepts a particle and maintains a space and
velocity distribution.  Meant to be used in boundaries with collect.

0.99	(PeterM ?)  Inception
0.991 (PeterM 8-29-95)  Add a method for computing Q, the sum of the
      charge which has hit the wall.
1.01  (JohnV 05-16-97) Add ftheta.
2.0   (JohnV 09-30-97) Revamped to store only f(x); should be fixed later.
2.1   (JohnV 06-28-98) Fixing later to include f(x,energy) and f(x,theta).

====================================================================
*/

#include "misc.h"
class Particle;
class Vector3;
class Species;

class PDistCol
{
	int nxbins;  // the number of 'bins' in the position;
	int nenergybins; // number of bins in energy;
	int nthetabins; // bins in theta;
	int nxpoints; // number of data points = nxbins+1;
	int nepoints;
	int ntpoints;
	Scalar x_min, x_max;
	Scalar energy_min, energy_max;
	Scalar theta_min, theta_max;
	int nsp;  /* the number of species */
	Scalar **fxdata;  /* [nsp][nxbins] */
	Scalar ***fedata; // [nsp][nxbins][nenergybins]
	Scalar ***ftdata; // [nsp][nxbins][nthetabins]
	Scalar xscale;
	Scalar escale;
	Scalar tscale;
	Scalar *x;  /* an array of the positions of the bins in x */
	Scalar *energy; // array of energy bins;
	Scalar *theta; // array of theta bins
	Scalar Qcol;  /* the sum of all the charge collected, ever */
	Scalar Qlast; /* the q from the previous timestep */
 public:
	void record_particle(Particle &p, int component);
	PDistCol(int nxbins, int nenergybins, int nthetabins, int nsp,
				Scalar energy_min, Scalar energy_max,
				Scalar x_min, Scalar x_max,
				Scalar theta_min, Scalar theta_max
				/*, Species *diag_species*/);
	~PDistCol();
	int *get_nxpoints() { return &nxpoints;};
	int *get_nepoints() { return &nepoints;};
	int *get_ntpoints() { return &ntpoints;};
	int get_nsp() { return nsp;};
	Scalar get_x_MIN() { return x_min;};
	Scalar get_x_MAX() { return x_max;};
	Scalar *get_x(){ return x;};
	Scalar *get_fxdata(int speciesID) {return fxdata[speciesID]; };
	Scalar get_energy_MIN() {return energy_min;};
	Scalar get_energy_MAX() {return energy_max;};
	Scalar *get_energy() {return energy;};
	Scalar **get_fedata(int speciesID) {return fedata[speciesID];};
	Scalar get_theta_MIN() {return theta_min;};
	Scalar get_theta_MAX() {return theta_max;};
	Scalar *get_theta() {return theta;};
	Scalar **get_ftdata(int speciesID) {return ftdata[speciesID];}
	Scalar getQ() {return Qcol;};
	Scalar getdQ() {
		Scalar dQ = Qcol - Qlast;
		Qlast = Qcol;
		return dQ;
	}
		
};

#ifdef oldcode
==============================================================

PDistCol::PDistCol(int _nxbins,	 int _nvbins, int _nthetabins, int _nsp,
						 Vector3 _vstart, Vector3 _vend,
						 Scalar _xstart,  Scalar _xend,
						 Scalar _theta1, Scalar _theta2,
						 Species *diag_species)
{ 
	int i,j;

	xend = _xend;
	xstart = _xstart;
	theta1 = _theta1*M_PI/180; // use rad internally
	theta2 = _theta2*M_PI/180;
	nxbins = _nxbins;
	nvbins = _nvbins;
	nthetabins = _nthetabins;
	nsp = _nsp;
	data = new Vector3 **[nsp];
	vstart = new Vector3 [nsp];
	vend = new Vector3 [nsp];
	vscale = new Vector3 [nsp];
	spf = new Species * [nsp];
	x = new Scalar [nxbins];
	v = new Vector3 *[nsp];
	if (nthetabins > 0) {
	  theta = new Scalar[nthetabins+1];
	  ftheta = new Scalar **[nsp];
	}

	Qlast = Qcol = 0;

	if(!nxbins) { nsp=0;return;};
	// we don't need to allocate diagnostics if nxbins==0

   int spec=diag_species->getID();
	vstart[spec]=_vstart;
	vend[spec]=_vend;
	dsp = spec;

	for(i=0;i<nsp;i++)
		{  spf[i] = 0;
			data[i]= new Vector3 *[nxbins];
			if (nthetabins > 0) ftheta[i] = new Scalar *[nxbins];
			v[i] = new Vector3[nvbins];
			for(j=0;j<nxbins;j++) {
				data[i][j] = new Vector3[nvbins];
				/* zero the arrays */
				memset(data[i][j],0,nvbins*sizeof(Vector3));
				if (nthetabins > 0) {
				  ftheta[i][j] = new Scalar[nthetabins];
				  memset(ftheta[i][j], 0, (nthetabins+1)*sizeof(Scalar));
				}
			}
			memset(v[i],0,nvbins*sizeof(Vector3));
		}

	vscale[spec] = Vector3(1.0,1.0,1.0).jvDivide((1.0/(Scalar)nvbins) * (vend[spec] - vstart[spec]));
	xscale = nxbins/(xend-xstart);
	spf[spec] = diag_species;
	for( i=0;i<nxbins;i++) x[i]= i/xscale + xstart;

  /*  set up the velocity array for species spec, the others must be done later */
	for(i=0;i<nvbins;i++) 
		v[spec][i] = Vector3((Scalar) i,(Scalar) i,(Scalar) i).jvDivide(vscale[spec]) +
			vstart[spec];

	if (nthetabins > 0) {
	  thetaScale = nthetabins/(theta2 - theta1);
	  for (i=0; i<=nthetabins; i++) theta[i] = i/thetaScale + theta1;
	}
}


/* component is which of the two positions to use for collection */
void PDistCol::record_particle(Particle &p, int component)
{
	int spec=p.get_speciesID();
	Scalar x = (component)? p.get_x().e2(): p.get_x().e1();
	int xi;  /*  the position in the space array */
	Vector3 vi;
	Vector3 *tmp;

   //collect the charge of this particle
	Qcol+=p.get_q();

   if(!nxbins) return;
	/* uninitialized vstart, vend, so we set them, assuming like temperatures. */
	if(!spf[spec]) 
		{ 
			spf[spec]=p.get_speciesPtr();
			Scalar cal_mass = spf[dsp]->get_m();
			vstart[spec]=vstart[dsp] * sqrt(cal_mass/spf[spec]->get_m());
			vend[spec]=vend[dsp] * sqrt(cal_mass/spf[spec]->get_m());
			vscale[spec] = Vector3(1.0,1.0,1.0).jvDivide((1.0/(Scalar)nvbins) 
												* (vend[spec] - vstart[spec]));
			for(int i=0;i<nvbins;i++) 
				v[spec][i] = Vector3((Scalar) i,(Scalar) i,(Scalar) i).jvDivide(
											vscale[spec]) +vstart[spec];

		}

	
	/* actually do the accumulation! */
	/* this assumes that the particle has the position on the boundary given by p*/
	xi = (int) ((x - xstart)*xscale);
	if(xi>=nxbins||xi<0) return;
	vi = p.get_u();

	// if nthetabins > 0 compute angular dist; currently ignore v3
	if (nthetabins > 0) {
	  Scalar theta;
	  if (component) theta = atan(vi.e2()/(vi.e1())+1e-20); // component == alongx2()
	  else theta = atan(vi.e1()/(vi.e2()+1e-20));
	  if (theta >= theta1 && theta <= theta2)
		 ftheta[spec][xi][(int) (0.5+thetaScale*(theta - theta1))] += p.get_np2c();
	}

	vi -= vstart[spec];
	vi = vi.jvMult(vscale[spec]) + Vector3(0.5,0.5,0.5);
	//if(vi.e1()>=nvbins||vi.e2()>=nvbins||vi.e3()>=nvbins||vi.e1()<0||vi.e2()<0||vi.e3()<0) return;

	/* store the recorded v */
	if(vi.e1()<nvbins && vi.e1() >= 0) {
		tmp = &data[spec][xi][(int)vi.e1()];
		tmp->set_e1(tmp->e1()+p.get_np2c());
	}
	if(vi.e2()<nvbins && vi.e2() >= 0) {
		tmp = &data[spec][xi][(int)vi.e2()];
		tmp->set_e2(tmp->e2()+p.get_np2c());
	}

	if(vi.e3()<nvbins && vi.e3() >= 0) {
		tmp = &data[spec][xi][(int)vi.e3()];
		tmp->set_e3(tmp->e3()+p.get_np2c());
	}
}

PDistCol::~PDistCol(void) {
	delete[] vstart;
	delete[] vend;
	delete[] vscale;
	delete[] spf;
	delete[] x;
	if (nthetabins > 0) delete[] theta;
	for(int i=0;i<nsp;i++)
		{
			delete[] v[i];
			for(int j=0;j<nxbins;j++) {
			  delete[] data[i][j];
			  if (nthetabins > 0) delete[] ftheta[i][j];
			}

			delete[] data[i];
			if (nthetabins > 0) delete[] ftheta[i];
		}
}			

#endif
