/*
=====================================================================
mcc.cpp

Methods for the MCC (Monte Carlo collision) class.

0.9	(JohnV 02-06-95) Initial draft.
1.01  (JohnV 01-02-97) Restructure for flexibility.
1.1   (JohnV 08-04-97) Added MCCPackages for Ar, Ne and Xe.
1.11  (JohnV 08-15-97) Repaired disparity between cross sections and
      thresholds for inelastic collisions.
1.12  (JohnV 10-01-97) Fixed mcc to work with supercycled timesteps.
1.13  (KC 10-20-97) Fixed mcc to work with M/m < 2.
1.14  (JohnV 02-11-99) Added He package (from Y. Ikeda)
2.5   (Bruhwiler 03-03-00) Added Lithium package (from S.M. Younger)
		
Consider how to make cross sections work for adjusted-mass particles
		
=====================================================================
*/

#include <stdio.h>
#include "mcc.h"
#include "species.h"
#include "maxwelln.h"
#include "particle.h"
#include "ptclgrp.h"

const Scalar	NperTORR =	8.3221e20;
const Scalar	MAX_ENERGY = 1000;

#define pow(x,y) pow((Scalar)(x),(Scalar)(y))

//-------------------------------------------------------------------
// CrossSection: provides basic cross section handling capability


//-------------------------------------------------------------------
//	eCrossSection methods

eCrossSection::eCrossSection(Scalar e0, Scalar e1, Scalar e2, Scalar sMax, Type _type)
{
	energy0 = e0;
	energy1 = e1;
	energy2 = e2;
	sigmaMax = sMax;
	const1 = sigmaMax/(energy1 - energy0);
	const3 = sigmaMax*energy2/log(energy2);
	type = _type;
}

Scalar eCrossSection::sigma(Scalar energy)
{
	if (energy <= energy0) return 0;
	if (energy <= energy1) return (energy - energy0)*const1;
	if (energy <= energy2) return sigmaMax;
	return const3*log(energy)/energy;
}


//-------------------------------------------------------------------
// Argon ionization cross section (Rapp & Golden) - energies in eV

Scalar argonIoniz::sigma(Scalar energy)
{										  // 
   if (energy < 15.76) return 0;
	else if (energy < 79) 
	  return 1.7155e-18*(energy-15.76)/(energy*energy)*log(0.0634*energy);
	else return 2.648e-18*(energy-15.76)/(energy*energy)*log(0.0344*energy);
}

//-------------------------------------------------------------------
//	iCrossSection methods

iCrossSection::iCrossSection(Scalar _a, Scalar _b, Type _type)
{
	a = _a;
	b = _b;
	type = _type;
}

Scalar iCrossSection::sigma(Scalar energy)
{
	return a + b/(sqrt(energy) + 1E-30);
}

//--------------------------------------------------------------------
// MCC: flexible collision package for handling multiple gases

MCC::~MCC()
{
  packageList.deleteAll(); // remove all list elements and data for packages
}

ParticleList& MCC::collide(ParticleGroupList& pgList)
{
  oopicListIter<MCCPackage> packageIter(packageList);
  for (packageIter.restart(); !packageIter.Done(); packageIter++)
	 packageIter()->collide(pgList, pList);
  return pList;
}

//----------------------------------------------------------------
// MCCPackage: individual gas packages

MCCPackage::MCCPackage(GAS_TYPE _type, Scalar p, Scalar _temp, Species* _eSpecies,
Species* _iSpecies, SpeciesList& sList, Scalar dt, int _ionzFraction, 
Scalar _ecxFactor, Scalar _icxFactor)
{
  type = _type;
  temp = _temp;
  gasDensity = NperTORR*p/(temp + 1e-30);
  eSpecies = _eSpecies;
  iSpecies = _iSpecies;
  ionzFraction = _ionzFraction;
  ecxFactor = _ecxFactor;
  icxFactor = _icxFactor;
  neCX = niCX = 0;
  eCX = iCX = NULL;

  addCrossSections();
}

void MCCPackage::init(SpeciesList& sList, Scalar dt)
{
  if (gasDensity > 0) {
	 extra = new Scalar[sList.nItems()];
	 collProb = new Scalar[sList.nItems()];
	 eEnergyScale = fabs(0.5/eSpecies->get_q_over_m());
	 iEnergyScale = fabs(0.5/iSpecies->get_q_over_m());

	 // check for electron and ion collisions
	 maxSigmaVe = maxSigmaVi = 0;
	 oopicListIter<Species> sIter(sList);
	 for (sIter.restart(); !sIter.Done(); sIter++){
		int id = sIter.current()->getID();
		extra[id] = 0.5;
		switch (sIter()->get_collisionModel()) {
		case none:
		  break;
		case boltzmann:
//			BoltzmannInit();
			break;
		case electron:
		case test:
		  if (maxSigmaVe == 0) maxSigmaVe = sumSigmaVe();
		  collProb[id] = 1 - exp(-gasDensity*maxSigmaVe*dt*sIter()->get_subcycleIndex()
										 /sIter()->get_supercycleIndex());
		  break;
		case ion:
		  if (maxSigmaVi == 0) maxSigmaVi = sumSigmaVi();
		  collProb[id] = 1 - exp(-gasDensity*maxSigmaVi*dt*sIter()->get_subcycleIndex()
										 /sIter()->get_supercycleIndex());
		  break;
		}
	 }

	 //	create isotropic maxwellian gas distribution with ion characteristics
	 gasDistribution = new Maxwellian(iSpecies);
	 gasDistribution->setThermal(temp, EV);
  }
  else extra = collProb = NULL;
}

MCCPackage::~MCCPackage()
{
  if (eCX) {
	 for (int i=0; i<neCX; i++) delete eCX[i];
	 delete[] eCX;
  }
  if (iCX) {
	 for (int i=0; i<niCX; i++) delete iCX[i];
	 delete[] iCX;
  }
  delete[] extra;
  delete[] collProb;
  delete gasDistribution;
}

void MCCPackage::addCrossSections()
{
  neCX = niCX = 0;
  eCX = iCX = NULL;
}

void MCCPackage::collide(ParticleGroupList& pgList, ParticleList& _pList)
{
	if (!pgList.head || !collProb) return;
	Species* species = pgList.head->data->get_species();
	CollisionModel collisionModel = species->get_collisionModel();
	if (collisionModel == none) return;
	int id = species->getID();
	if (collProb[id] < 1e-20) return;
	pList = &_pList;
	int n = 0;
	oopicListIter<ParticleGroup> pgIter(pgList);
	for (pgIter.restart(); !pgIter.Done(); pgIter++) 
		n += pgIter()->get_n();
	if (n==0) return;
	extra[id] += n*collProb[id];
	int nCollisions = (int)extra[id];
	extra[id] -= nCollisions;
	while (nCollisions--){
	  index = (int)(n*frand());			//	get extended particle index
	  for (pgIter.restart(); !pgIter.Done(); pgIter++) // find it
			if (index < pgIter()->get_n()) break;
			else index -= (int)pgIter()->get_n();
	  pg = pgIter();
	  u = pg->get_u(index);
	  BOOL relativistic = FALSE;
	  v = u.magnitude() + 1E-30;
	  if (v > 1e8) { 	// remove gamma
			relativistic = TRUE;
			u /= pg->gamma(index);
			v = u.magnitude() + 1E-30;
	  }
	  energy = pg->energy_eV(index); // in eV
	  Scalar	random = frand()/v;
	  Scalar	sumSigma = 0;
	  if ((collisionModel == electron)||(collisionModel == test)) {
			random *= maxSigmaVe;
			// loop over all possible collisions with this gas
			for (int i=0; i<neCX; i++) {
				sumSigma += ecxFactor*eCX[i]->sigma(energy);
				if (random <= sumSigma) {
					dynamic(*eCX[i]);
					break; // leave for loop if collision done
				}
			}
	  }
	  else if (collisionModel == ion) {
			random *= maxSigmaVi;
			// loop over all possible collisions with this gas
			for (int i=0; i<niCX; i++) {
				sumSigma += icxFactor*iCX[i]->sigma(energy);
				if (random <= sumSigma) {
					dynamic(*iCX[i]);
					break; // leave for loop if collision done
				}
			}
		}
/*		else if (collisionModel == test) {
			random *= maxSigmaVe;
			// loop over all possible collisions with this gas
			for (int i=0; i<(ntCX-1); i++) {
				sumSigma += ecxFactor*tCX[i]->sigma(energy);
				if (random <= sumSigma) {
					dynamic(*tCX[i]);
					break; // leave for loop if collision done
				}
				}
				}
*/ 
// This isn't needed any more
	  if (relativistic) u *= pg->gamma(u);
	  Vector3 *uPtr = pg->get_u();
	  uPtr[index] = u;
	}
}

// compute null (max) cross section for e-neutral collisions
Scalar MCCPackage::sumSigmaVe()
{
  maxSigmaVe = 0;
  for (Scalar energy=0; energy<MAX_ENERGY; energy+=0.5) {
	 Scalar temp = 0;
	 for (int i=0; i<neCX; i++) temp += eCX[i]->sigma(energy);
	 maxSigmaVe = MAX(maxSigmaVe, sqrt(energy)*temp);
  }
  maxSigmaVe *= ecxFactor/sqrt(eEnergyScale);
  return maxSigmaVe;
}

// compute null (max) cross section for ion-neutral collisions
Scalar MCCPackage::sumSigmaVi()
{
  maxSigmaVi = 0;
  for (Scalar energy=0; energy<MAX_ENERGY; energy+=0.5) {
	 Scalar temp = 0;
	 for (int i=0; i<niCX; i++) temp += iCX[i]->sigma(energy);
	 maxSigmaVi = MAX(maxSigmaVi, sqrt(energy)*temp);
  }
  maxSigmaVi *= icxFactor/sqrt(iEnergyScale);
  return maxSigmaVi;
}

void MCCPackage::dynamic(CrossSection& cx)
{
  switch(cx.get_type()) 
	 {
	 case CrossSection::ELASTIC:
		elastic(cx);
		break;
	 case CrossSection::EXCITATION:
		excitation(cx);
		break;
	 case CrossSection::IONIZATION:
		ionization(cx);
		break;
	 case CrossSection::ION_ELASTIC:
		ionElastic(cx);
		break;
	 case CrossSection::CHARGE_EXCHANGE:
		chargeExchange(cx);
		break;
	 default:
		printf("unrecognized collision type!\n");
	 }
}

void MCCPackage::elastic(CrossSection& cx)
{
  u /= v; // normalize
  newVelocity(energy, v, u, TRUE); // scatter electron
}

void MCCPackage::excitation(CrossSection& cx)
{
  u /= v; // is this in the right place? Should be after energy loss?
  energy -= cx.get_threshold(); // subtract inelastic energy loss
  v = sqrt(energy/eEnergyScale); // recompute magnitude of velocity
  newVelocity(energy, v, u, TRUE);
}

void MCCPackage::ionization(CrossSection& cx)
{
  u /= v;
  energy -= cx.get_threshold(); // subtract ionization energy
  Scalar rEnergy = 10*tan(frand()*atan(energy/20));
  energy -= rEnergy;			//	partition remaining energy
				
  //	scatter created electron:
  v = sqrt(rEnergy/eEnergyScale);
  Vector3 uNew = u;
  newVelocity(rEnergy, v, uNew, FALSE);
  pList->add(new Particle(pg->get_x(index), uNew*pg->gamma(uNew), eSpecies, 
								  pg->get_np2c(index), pg->isVariableWeight()));

  // create n=ionzFraction ions from isotropic gas distribution
  for (int i=0; i<ionzFraction; i++) {
	 uNew = gasDistribution->get_V();
	 pList->add(new Particle(pg->get_x(index), uNew, iSpecies,
		 pg->get_np2c(index)/ionzFraction, pg->isVariableWeight()));
  }

  // scatter incident electron

  v = sqrt(energy/eEnergyScale);
  newVelocity(energy, v, u, FALSE);
}

void MCCPackage::ionElastic(CrossSection& cx)
{
  Vector3 gasVelocity = gasDistribution->get_V();
  u -= gasVelocity;					//	center of mass frame

  Scalar up1, up2, up3, mag;
  Scalar r11, r12, r13, r21, r22, r23, r31, r32, r33;
			
  Scalar coschi= sqrt(frand());
  Scalar sinchi = sqrt(1. -coschi*coschi);
	
  Scalar phi1  = TWOPI*frand();
  Scalar cosphi = cos(phi1);
  Scalar sinphi = sin(phi1);
	
  r13 = u.e1()/v;
  r23 = u.e2()/v;
  r33 = u.e3()/v;
	
  if(r33 == 1.0) { up1= 0;  up2= 1;  up3= 0; }
  else           { up1= 0;  up2= 0;  up3= 1; }
  
  r12 = r23*up3 -r33*up2;
  r22 = r33*up1 -r13*up3;
  r32 = r13*up2 -r23*up1;
  mag = sqrt(r12*r12 + r22*r22 + r32*r32) + 1E-30;
	
  r12/= mag;
  r22/= mag;
  r32/= mag;
	
  r11 = r22*r33 -r32*r23;
  r21 = r32*r13 -r12*r33;
  r31 = r12*r23 -r22*r13;
	
  u.set_e1(v*coschi*(r11*sinchi*cosphi +r12*sinchi*sinphi +r13*coschi));
  u.set_e2(v*coschi*(r21*sinchi*cosphi +r22*sinchi*sinphi +r23*coschi)); 
  u.set_e3(v*coschi*(r31*sinchi*cosphi +r32*sinchi*sinphi +r33*coschi));
  u += gasVelocity;					//	back to lab frame
}

void MCCPackage::chargeExchange(CrossSection& cx)
{
  Vector3 gasVelocity = gasDistribution->get_V();
  u -= gasVelocity;					//	center of mass frame
  u.set_e1(0);
  u.set_e2(0);
  u.set_e3(0);
  u += gasVelocity;					//	back to lab frame
}

void MCCPackage::newVelocity(Scalar energy, Scalar v, Vector3& u, int eFlag)
{
	Scalar	up1, up2, up3;
	Scalar	coschi;
	Scalar	sinchi;
	if (energy < 1e-30) coschi = 1.0;
	else coschi = 1 + 2.0*(1.0 - pow(energy + 1.0, frand()))/energy;
	sinchi = sqrt(1 - coschi*coschi);
	Scalar	phi = TWOPI*frand();
	Scalar	cosphi = cos(phi);
	Scalar	sinphi = sin(phi);

	// transfer momentum:

//	if (eFlag) v *= sqrt(1 - 2*eSpecies->get_m()/iSpecies->get_m()*(1 - coschi));
	// doesn't work when m/M>.5

	Scalar me = eSpecies->get_m();
	Scalar mi = iSpecies->get_m();
	v*= sqrt(1 - 2*me*mi/(me+mi)/(me+mi)*(1-coschi));

	Scalar	r13 = u.e1();
	Scalar	r23 = u.e2();
	Scalar	r33 = u.e3();

	if (r33 == 1)	{up1 = 0; up2 = 1; up3 = 0;}
	else				{up1 = 0; up2 = 0; up3 = 1;}

	Scalar	r12 = r23*up3 - r33*up2;
	Scalar	r22 = r33*up1 - r13*up3;
	Scalar	r32 = r13*up2 -r23*up1;
	Scalar	mag = sqrt(r12*r12 + r22*r22 + r32*r32);
  
	r12 /= mag;
	r22 /= mag;
	r32 /= mag;
  
	Scalar	r11 = r22*r33 - r32*r23;
	Scalar	r21 = r32*r13 - r12*r33;
	Scalar	r31 = r12*r23 - r22*r13;

	u.set_e1(v*(r11*sinchi*cosphi + r12*sinchi*sinphi + r13*coschi));
	u.set_e2(v*(r21*sinchi*cosphi + r22*sinchi*sinphi + r23*coschi)); 
	u.set_e3(v*(r31*sinchi*cosphi + r32*sinchi*sinphi + r33*coschi));
}

//-------------------------------------------------------------
// ArMCC: argon MCC package

ArMCC::ArMCC(Scalar p, Scalar temp, Species* eSpecies, Species* iSpecies, 
SpeciesList& sList, Scalar dt, int ionzFraction, Scalar ecxFactor, Scalar icxFactor)
 : MCCPackage(Ar, p, temp, eSpecies, iSpecies, sList, dt, ionzFraction, ecxFactor, 
icxFactor)
{
  addCrossSections();
  init(sList, dt);
}

void ArMCC::addCrossSections()
{
  neCX = 3; // electron model
  eCX = new CrossSection*[neCX];
  eCX[0] = new eCrossSection(0.3, 15, 20, 1.2e-19, CrossSection::ELASTIC);
  eCX[1] = new eCrossSection(11.55, 30, 100, 7e-21, CrossSection::EXCITATION);
  eCX[2] = new FuncCrossSection(sigmaIz, 15.76, CrossSection::IONIZATION);
  //	ionizationCX = new eCrossSection(15.76, 30, 100, 3e-20, IONIZATION);

  niCX = 2; // ion model
  iCX = new CrossSection*[niCX];
  iCX[0] = new iCrossSection(1.8e-19, 4e-19, CrossSection::ION_ELASTIC);
  iCX[1] = new iCrossSection(2e-19, 5.5e-19, CrossSection::CHARGE_EXCHANGE);
}

//-------------------------------------------------------------------
// Argon ionization cross section (Rapp & Golden) - energies in eV

Scalar ArMCC::sigmaIz(Scalar energy)
{
   if (energy < 15.76) return 0;
	else if (energy < 79) 
	  return 1.7155e-18*(energy-15.76)/(energy*energy)*log(0.0634*energy);
	else return 2.648e-18*(energy-15.76)/(energy*energy)*log(0.0344*energy);
}

//-------------------------------------------------------------
// NeMCC: neon MCC package

NeMCC::NeMCC(Scalar p, Scalar temp, Species* eSpecies, Species* iSpecies, 
SpeciesList& sList, Scalar dt, int ionzFraction, Scalar ecxFactor, Scalar icxFactor)
 : MCCPackage(Ne, p, temp, eSpecies, iSpecies, sList, dt, ionzFraction, ecxFactor, 
icxFactor)
{
  addCrossSections();
  init(sList, dt);
}

void NeMCC::addCrossSections()
{
  neCX = 3; // electron model
  eCX = new CrossSection*[neCX];
  eCX[0] = new FuncCrossSection(sigmaElastic, 0, CrossSection::ELASTIC);
  eCX[1] = new FuncCrossSection(sigmaExc, 15.8489, CrossSection::EXCITATION);
  eCX[2] = new FuncCrossSection(sigmaIz, 20.893, CrossSection::IONIZATION);

  niCX = 1; // ion model
  iCX = new CrossSection*[niCX];
  iCX[0] = new iCrossSection(0, 5.98e-20, CrossSection::CHARGE_EXCHANGE);
}

Scalar NeMCC::sigmaElastic(Scalar energy)
{
  if (energy < 2.63027) return 1.27808e-20 + energy*(4.22329e-21 - 7.31854e-22*energy);
  else if (energy < 19.4984) return 1.66333e-20 + energy*(8.66972e-22 - 1.14153e-23*energy);
  else if (energy < 33.8844) return 2.2297e-20 + energy*(5.8066e-22 - 1.18023e-23*energy);
  else if (energy < 190.546) return 3.862e-20 + energy*(-3.19062e-22 + 8.00046e-25*energy);
  else if (energy < 501.187) return 1.37254e-20 + energy*(-4.66403e-23 + 4.6479e-26*energy);
  else if (energy < 1000) return 4.38798e-21 + energy*(-6.19168e-24 + 2.5747e-27*energy);
  else return 0;
}

Scalar NeMCC::sigmaExc(Scalar energy) // summed excitations
{
  if (energy <= 15.8489) return 0;
  else if (energy < 23.9883) return -2.37959e-22 + energy*(-5.87101e-23 + 4.6417e-24*energy);
  else if (energy < 36.3078) return -3.49005e-21 + energy*(2.76934e-22 - 3.69893e-24*energy);
  else if (energy < 54.9541) return 2.75471e-22 + energy*(6.0912e-23 - 5.99808e-25*energy);
  else if (energy < 251.189) return 2.271e-21 + energy*(-8.78552e-24 + 1.41951e-26*energy);
  else if (energy < 616.595) return 1.53078e-21 + energy*(-2.7892e-24 + 1.90737e-27*energy);
  else if (energy < 1000) return 1.05134e-21 + energy*(-1.07925e-24 + 3.83108e-28*energy);
  else return 0;
}

Scalar NeMCC::sigmaIz(Scalar energy)
{
  if (energy <= 20.893) return 0;
  else if (energy < 54.9541) return -2.34018e-21 + energy*1.10755e-22;
  else if (energy < 125.893) return -2.4556e-21 + energy*(1.42218e-22 - 5.37266e-25*energy);
  else if (energy < 190.546) return 2.62391e-21 + energy*(5.39463e-23 - 1.54319e-25*energy);
  else if (energy < 436.516) return 8.04001e-21 + energy*(-2.68853e-24 - 6.11372e-27*energy);
  else if (energy < 1000) return 9.16045e-21 + energy*(-9.59021e-24 + 3.85096e-27*energy);
  else return 0;
}

//-------------------------------------------------------------
// XeMCC: xenon MCC package

XeMCC::XeMCC(Scalar p, Scalar temp, Species* eSpecies, Species* iSpecies, 
SpeciesList& sList, Scalar dt, int ionzFraction, Scalar ecxFactor, Scalar icxFactor)
 : MCCPackage(Xe, p, temp, eSpecies, iSpecies, sList, dt, ionzFraction, ecxFactor, 
icxFactor)
{
  addCrossSections();
  init(sList, dt);
}

void XeMCC::addCrossSections()
{
  neCX = 3; // electron model
  eCX = new CrossSection*[neCX];
  eCX[0] = new FuncCrossSection(sigmaElastic, 0, CrossSection::ELASTIC);
  eCX[1] = new FuncCrossSection(sigmaExc, 7.94328, CrossSection::EXCITATION);
  eCX[2] = new FuncCrossSection(sigmaIz, 11.2202, CrossSection::IONIZATION);

  niCX = 1; // ion model
  iCX = new CrossSection*[niCX];
  iCX[0] = new iCrossSection(0, 4.541e-20, CrossSection::CHARGE_EXCHANGE);
}

Scalar XeMCC::sigmaElastic(Scalar energy)
{
  if (energy <= 1) return 0;
  else if (energy < 3.46737) return -2.56426e-20 + energy*(3.05131e-20 + 1.01401e-20*energy);
  else if (energy < 4.2658) return -7.02412e-19 + energy*(4.15012e-19 - 4.46826e-20*energy);
  else if (energy < 5.62341) return -3.14744e-19 + energy*(2.19764e-19 - 2.027176e-20*energy);
  else if (energy < 12.0226) return 2.23972e-19 + energy*(2.61272e-19 - 2.81077e-21*energy);
  else if (energy < 25.704) return 3.39014e-19 + energy*(-2.28038e-20 + 4.64864e-22*energy);
  else if (energy < 72.4436) return 9.50404e-20 + energy*(-1.80585e-21 + 1.24365e-23*energy);
  else if (energy < 190.546) return 4.51567e-20 + energy*(-2.78976e-22 + 6.93189e-25*energy);
  else if (energy < 537.032) return 2.59891e-20 + energy*(-5.72321e-23 + 4.79915e-26*energy);
  else if (energy < 1000) return 1.46698e-20 + energy*(-1.28135e-23 + 3.98769e-27*energy);
  else return 0;
}

Scalar XeMCC::sigmaExc(Scalar energy) // summed excitations
{
  if (energy <= 7.94328) return 0;
  else if (energy < 11.2202) return 7.71935e-20 + energy*(-1.8566e-20 + 1.117e-21*energy);
  else if (energy < 20.893) return -5.72625e-20 + energy*(8.29355e-21 - 2.02235e-22*energy);
  else if (energy < 33.8844) return 9.4784e-21 + energy*(1.36322e-21 - 2.20758e-23*energy);
  else if (energy < 190.546) return 3.66375e-20 + energy*(-1.87228e-22 + 3.94988e-25*energy);
  else if (energy < 467.735) return 2.51242e-20 + energy*(-6.45713e-23 + 6.14246e-26*energy);
  else if (energy < 1000) return 1.4849e-20 + energy*(-1.78916e-23 + 7.74732e-27*energy);
  else return 0;
}

Scalar XeMCC::sigmaIz(Scalar energy)
{
  if (energy <= 11.2202) return 0;
  else if (energy < 44.6684) return -4.78153e-20 + energy*(4.75275e-21 - 5.83101e-23*energy);
  else if (energy < 67.6083) return 4.90931e-20 + energy*(-6.99912e-23 + 1.11671e-24*energy);
  else if (energy < 102.329) return 3.73779e-20 + energy*(3.25673e-22 - 2.15897e-24*energy);
  else if (energy < 204.174) return 6.20189e-20 + energy*(-1.62295e-22 + 2.70937e-25*energy);
  else if (energy < 575.44) return 4.1e-20 - 3.49995e-24*energy;
  else if (energy < 812.831) return 2.19858e-20 + energy*(9.27581e-23 - 1.09139e-25*energy);
  else if (energy < 1000) return 8.37604e-20 + energy*(-1.01201e-22 + 3.62937e-26*energy);
  else return 0;
}

//-------------------------------------------------------------
// HMCC:  Hydrogen MCC package

HMCC::HMCC(Scalar p, Scalar temp, Species* eSpecies, Species* iSpecies, 
SpeciesList& sList, Scalar dt, int ionzFraction, Scalar ecxFactor, Scalar icxFactor)
 : MCCPackage(Ar, p, temp, eSpecies, iSpecies, sList, dt, ionzFraction, ecxFactor, 
icxFactor)
{
  addCrossSections();
  init(sList, dt);
}

void HMCC::addCrossSections()
{
  neCX = 1; // electron model
  eCX = new CrossSection*[neCX];
  eCX[0] =  new FuncCrossSection(sigmaIz, 13.6, CrossSection::IONIZATION);

}

Scalar HMCC::sigmaIz(Scalar energy)
{
  if (energy <= 13.6) return 0;
  else if (energy <= 17.789) return 8.208063e-21*log(energy)-2.071614e-20;
  else if (energy <= 53.6999) return 3.751578e-21*log(energy)-7.887766e-21;
  else return (2.996231e-19*log(energy)-8.145988e-19)/energy;
}
//-------------------------------------------------------------------



//-------------------------------------------------------------
// LiMCC: Lithium MCC package    -- added 03/02/00 by Bruhwiler

LiMCC::LiMCC(Scalar p, Scalar temp, Species* eSpecies, Species* iSpecies, 
       SpeciesList& sList, Scalar dt, int ionzFraction, Scalar ecxFactor,
       Scalar icxFactor)
 : MCCPackage(Li, p, temp, eSpecies, iSpecies, sList, dt,
                       ionzFraction, ecxFactor, icxFactor) {
  threshold = 5.392;
  addCrossSections();
  init(sList, dt);
}

void LiMCC::addCrossSections() {
    // specify the number of electron collision models that are implemented
  neCX = 1;
    // create an array of CrossSection pointers
  eCX = new CrossSection*[neCX];
    // load up the array of CrossSection pointers
  eCX[0] = new LithiumIonizCrossSection();
}


/*  Method for calculating e-impact ionization cross-sections
 *  using the fitting function from Eq. 4 of the 1981 J.Q.S.R.T.
 *  paper by S.M. Younger.
 *
 *  Units are eV for energy and cm^2 for cross-sections.
 *  First added March 2, 2000, by David Bruhwiler
 */
Scalar LiMCC::sigmaIz(Scalar energy) {

// This is specific for plasma accelerator simulations!!!
// Needs to be generalized to use flag from input file!!!
//  if (energy < 50.) return 0.;

// normalized impact energy
// -- assumed that both energy and threshold are in eV
  Scalar U = energy / 5.392;
  if ( U <= 1. ) return 0.;

  Scalar oneMinusUinv = 1. - 1. / U;
  Scalar logU = log(U);
  Scalar value = 1.35e-13 * oneMinusUinv
               - 6.51e-14 * oneMinusUinv * oneMinusUinv
               + 5.99e-15 * logU
               - 8.49e-14 * logU / U;

  return value / ( energy * 5.392 );
}


void MCCPackage::BoltzmannInit()
{

	const int NumberBins=100;
//	const int IS; //IntegrationSteps
	const Scalar Tmin = .05;  // this needs to match min temp in boltzman.cpp
	const Scalar Tmax = 10;  // This is hopefull the hightest temperture.
	const Scalar EnergyStep=.5;

	Scalar T;
	G = new Scalar[NumberBins];
	E = new Scalar[NumberBins];
	for (int n=0; n<NumberBins; n++){
		G[n] = 0;
		E[n] = 0;
	}
	
	Scalar temp = 0;
	
	for (int i=0; i<neCX; i++)
		switch (eCX[i]->get_type())
			{
			case CrossSection::IONIZATION: {
				for (int n=0; n<NumberBins; n++){
					T = n*(Tmax-Tmin)/(NumberBins-1)+Tmin;
					for (Scalar energy=EnergyStep/2; energy<MAX_ENERGY; energy+=EnergyStep){
						if (energy/T<50)
							temp += energy*exp(-energy/T)*eCX[i]->sigma(energy);
					}
					temp *= gasDensity*sqrt(2/(T*M_PI*eSpecies->get_m()))*2/T;
					G[n] += temp;
					E[n] -= temp*eCX[i]->get_threshold();
				}
				break;
			}
			case CrossSection::ELASTIC:
				{
				for (int n=0; n<NumberBins; n++){
					T = n*(Tmax-Tmin)/(NumberBins-1)+Tmin;
					for (Scalar energy=EnergyStep/2; energy<MAX_ENERGY; energy+=EnergyStep){
						if (energy/T<50)
							temp += energy*exp(-energy/T)*eCX[i]->sigma(energy);
					}
					temp *= gasDensity*sqrt(2/(T*M_PI*eSpecies->get_m()))*2/T;
					E[n] -= temp*eCX[i]->get_threshold();
				}
				break;
			}
			case CrossSection::EXCITATION:
				{
				for (int n=0; n<NumberBins; n++){
					T = n*(Tmax-Tmin)/(NumberBins-1)+Tmin;
					for (Scalar energy=EnergyStep/2; energy<MAX_ENERGY; energy+=EnergyStep){
						if (energy/T<50)
							temp += energy*exp(-energy/T)*eCX[i]->sigma(energy);
					}
					temp *= gasDensity*sqrt(2/(T*M_PI*eSpecies->get_m()))*2/T;
					E[n] -= temp*eCX[i]->get_threshold();
				}
				break;
			}
			default:
				printf("unrecognized collision type!\n");
			}
	
}


/*  Method for calculating e-impact ionization cross-sections
 *  using the fitting function from Eq. 4 of the 1981 J.Q.S.R.T.
 *  paper by S.M. Younger.
 *
 *  Units are eV for energy and cm^2 for cross-sections.
 *  First added March 2, 2000, by David Bruhwiler
 */
Scalar LithiumIonizCrossSection::sigma(Scalar energy) {

// This is specific for plasma accelerator simulations!!!
// Needs to be generalized to use flag from input file!!!
  if (energy < 800.) return 0.;

	// normalized impact energy
	// -- assumed that both energy and threshold are in eV
  Scalar U = energy / threshold;
  if ( U <= 1. ) return 0.;

  Scalar oneMinusUinv = 1. - 1. / U;
  Scalar logU = log(U);
  Scalar value = youngerParamA * oneMinusUinv
               + youngerParamB * oneMinusUinv * oneMinusUinv
               + youngerParamC * logU
               + youngerParamD * logU / U;

	return value / ( energy * threshold );
}
