/*
=====================================================================
mcc.h

Describes the MCC (Monte Carlo collision) class.  This class provides 
collisions by operating on ParticleGroups.

0.9	(JohnV 02-06-95) Initial draft.
1.1 (JohnV 12-12-96) Added Rapp&Golden Argon ionization cross section
2.5 (Bruhwiler 03-01-00) Added electron impact ionization cross section for Li
		
=====================================================================
*/

#ifndef	__MCC_H
#define	__MCC_H

#include "ovector.h"
#include "misc.h"
#include "particle.h"

class Maxwellian;
class	Species;

class CrossSection
{
  void newVelocity(Scalar energy, Scalar v, Vector3& u, int eFlag);
public:
  enum Type {ELASTIC, EXCITATION, IONIZATION, ION_ELASTIC, CHARGE_EXCHANGE, TEST_TO_BOLTZMANN} type;
  virtual Scalar sigma(Scalar energy)=0;
  virtual Scalar get_threshold() {return 0;} // return inelastic threshold energy
  Type get_type() {return type;}
};

class eCrossSection : public CrossSection
{
	Scalar	energy0;							//	(eV) threshold
	Scalar	energy1;							//	(eV) start of plateau
	Scalar	energy2;							//	(eV) begin logE/E drop 
	Scalar	sigmaMax;						//	peak cross section m^2
	Scalar	const1;							//	internal constant for energy range 1
	Scalar	const3;							//	internal constant for energy range 3
public:
	eCrossSection(Scalar e0, Scalar e1, Scalar e2, Scalar sMax, Type type);
	virtual Scalar sigma(Scalar energy);
	//Scalar	get_energy0() {return energy0;}
	virtual Scalar get_threshold() {return energy0;}
};

class FuncCrossSection : public CrossSection
{
protected:
   Scalar threshold;
   Scalar (*sigmaPtr)(Scalar energy);
public:
	 FuncCrossSection(Scalar (*sigma)(Scalar), Scalar _threshold, Type _type)
	  {sigmaPtr = sigma; threshold = _threshold; type = _type;}
	virtual Scalar sigma(Scalar energy) {return (*sigmaPtr)(energy);}
	virtual Scalar get_threshold() {return threshold;}
};


class argonIoniz : public eCrossSection
{
 public:
   argonIoniz() : eCrossSection(15.76, 30, 100, 1, IONIZATION) {};
   virtual Scalar sigma(Scalar energy);
};

class iCrossSection : public CrossSection
{
	Scalar	a;									//	cross section coeff.
	Scalar	b;									// ditto
public:
	iCrossSection(Scalar a, Scalar b, Type type);
	virtual Scalar	sigma(Scalar energy);
};

class MCCPackage
{
public:
	enum GAS_TYPE {Ar, Ne, Xe, H, Li} type;
 private:
   Scalar temp;							//	gas temp in eV
	Scalar gasDensity;						//	gas number density m^-3
	Maxwellian* gasDistribution;
//	Scalar MinPot;
	Scalar* collProb;
	Scalar ecxFactor;                    // factor to multiply by electron cx
	Scalar icxFactor;                    // factor to multiply by ion cx
	Scalar maxSigmaVe;						// MAX(sigma*v)
	Scalar maxSigmaVi;
	Scalar* extra;                      // fractional collisions from previous steps
	int ionzFraction; // fraction of particles to create in ionization event
	Species* eSpecies; // created in ionization
	Species* iSpecies; // created in ionization
	Scalar	eEnergyScale;					//	m/2e (converts (m/s)^2 -> eV)
	Scalar	iEnergyScale;					//	M/2e
	void	newVelocity(Scalar energy, Scalar v, Vector3& u, int eFlag);
	Scalar sumSigmaVe();                // returns MAX(sigma*ve)
	Scalar sumSigmaVi();                // returns MAX(sigma*vi)
	// following are used in both collision() and elastic(), excitation(), etc.
	Scalar v;
	Vector3 u;
	Scalar energy;
	ParticleList* pList;
	ParticleGroup* pg;
	int index;
	virtual void addCrossSections();
public:
	MCCPackage(GAS_TYPE type, Scalar p, Scalar temp, Species* eSpecies,
		Species* iSpecies, SpeciesList& sList, 
						 Scalar dt, int ionzFraction, Scalar ecxFactor, Scalar icxFactor);
	virtual ~MCCPackage();
	virtual void init(SpeciesList& sList, Scalar dt);
	virtual void BoltzmannInit();
	Scalar *G,*E;
	int neCX; // number of cross sections
	int niCX; // number of cross sections
	int ntCX; 
	int nbCX;
	CrossSection** eCX;
	CrossSection** iCX;
	CrossSection** tCX;
	CrossSection** bCX;
	void collide(ParticleGroupList &pgList, ParticleList& pList);
	virtual void dynamic(CrossSection& cx);
	virtual void elastic(CrossSection& cx);
	virtual void excitation(CrossSection& cx);
	virtual void ionization(CrossSection& cx);
	virtual void ionElastic(CrossSection& cx);
	virtual void chargeExchange(CrossSection& cx);
};

class ArMCC : public MCCPackage
{
  static Scalar sigmaIz(Scalar energy);
public:
  ArMCC(Scalar p, Scalar temp, Species* eSpecies, Species* iSpecies,  
				SpeciesList& sList,
				Scalar dt, int ionzFraction, Scalar ecxFactor, Scalar icxFactor);
  virtual void addCrossSections();
};

class NeMCC : public MCCPackage
{
  static Scalar sigmaElastic(Scalar energy);
  static Scalar sigmaExc(Scalar energy);
  static Scalar sigmaIz(Scalar energy);
public:
  NeMCC(Scalar p, Scalar temp, Species* eSpecies, Species* iSpecies,  
				SpeciesList& sList,
				Scalar dt, int ionzFraction, Scalar ecxFactor, Scalar icxFactor);
  virtual void addCrossSections();
};

class XeMCC : public MCCPackage
{
  static Scalar sigmaElastic(Scalar energy);
  static Scalar sigmaExc(Scalar energy);
  static Scalar sigmaIz(Scalar energy);
public:
  XeMCC(Scalar p, Scalar temp, Species* eSpecies, Species* iSpecies, 
				SpeciesList& sList,
				Scalar dt, int ionzFraction, Scalar ecxFactor, Scalar icxFactor);

  virtual void addCrossSections();
};


/** Monte Carlo collision (MCC) class for Lithium.
 *  Only electron impact ionization is implemented at present.
 *  First added March 2, 2000.
 *
 * @author David Bruhwiler
 */
class LiMCC : public MCCPackage {

  public:

    /** Constructor
     */
    LiMCC(Scalar p, Scalar temp, Species* eSpecies, Species* iSpecies, 
				SpeciesList& sList,
				Scalar dt, int ionzFraction, Scalar ecxFactor, Scalar icxFactor);

    /** Method from MCCPackage class for adding cross sections.
     */
    virtual void addCrossSections();

  protected:

    /** Static method that calculates the ionization cross-section in cm^2
		 *  as a function of the electron impact energy in eV.
     */
    static Scalar sigmaIz(Scalar energy);

    /** Ionization threshold energy in eV.
     */
    Scalar threshold;
};


class MCC
{
  oopicList<MCCPackage> packageList;
  ParticleList	pList;					//	list of newly created particles
public:
  MCC() {}
  ~MCC();
  void addPackage(MCCPackage *p) {packageList.add(p);}
  ParticleList& collide(ParticleGroupList& pgList);
};

class HMCC : public MCCPackage
{
	static Scalar sigmaIz(Scalar energy);
public:
  HMCC(Scalar p, Scalar temp, Species* eSpecies, Species* iSpecies,  
				SpeciesList& sList,
				Scalar dt, int ionzFraction, Scalar ecxFactor, Scalar icxFactor);
  virtual void addCrossSections();
};



/** Class for calculating e-impact ionization cross-sections for
 *  Lithium.  The Younger fitting parameters come directly from
 *  Journal of Quantitative Spectroscopy and Radiative Transfer,
 *  Vol 26, 1981, p. 329,  by S.M. Younger.  The paper by Bray,
 *  "Calculation of electron-impact ionization of lithium-like
 *  targets," J. Phys. B: At. Mol. Opt. Phys., Vol. 28 (1995),
 *  p. L247,  suggests that this fit might yield a cross-section
 *  that's too large for impact energies less than 30 eV.
 *
 *  This is for ionization of neutral Lithium, assuming the electron
 *  is removed from the 2s shell (ie outer electron).  This dominates
 *  the cross-section for removal from the inner shell.
 *
 *  Units are eV for energy and cm^2 for cross-sections.
 *  First added March 2, 2000.
 *
 * @author David Bruhwiler
 */
class LithiumIonizCrossSection : public CrossSection {

  public:

    /** Constructor.
     */
    LithiumIonizCrossSection() {
      type = IONIZATION;
      threshold = 5.392;         // ionization threshold in eV
      youngerParamA =  1.35e-13;
      youngerParamB = -6.51e-14;
      youngerParamC =  5.99e-15;
      youngerParamD = -8.49e-14;
    }

    /** Override the virtual method that returns cross-section in cm^2 as a
		 *  function of the electron impact energy in eV.
     */
    virtual Scalar sigma(Scalar energy);

    /** Override virtual method that returns ionization threshold in eV.
     */
    virtual Scalar get_threshold() {return threshold;}

  private:

    /** Ionization threshold energy in eV.
     */
    Scalar threshold;

    /** Younger fitting parameter A, in units of cm^2 * eV^2
     */
    Scalar youngerParamA;

    /** Younger fitting parameter B, in units of cm^2 * eV^2
     */
    Scalar youngerParamB;

    /** Younger fitting parameter C, in units of cm^2 * eV^2
     */
    Scalar youngerParamC;

    /** Younger fitting parameter D, in units of cm^2 * eV^2
     */
    Scalar youngerParamD;
};

#endif	//	ifndef __MCC_H

