/*
=====================================================================
plsmadev.cpp

Provides methods for the PlasmaDevice class.

0.9	(JohnV 10-16-94) Initial draft to hook up to advisor.  This
		version replaces the original version completely.
0.911	(JohnV 11-23-94) Aesthetics.
0.92	(JohnV 11-28-94) Add deleteSpatialRegions().
0.921	(JohnV 11-30-94) Add PlasmaDevice(SpatialRegionList*).

(Hadon 12-17-94) replace AllPGGroupListIter with ScalarIter
(Hadon 12-19-94) change typedef AllVectorFeilds to ScalarIter
(Hadon 12-19-94) Remove getBField (not needed by new ScalarIter's)
(Hadon & James 3-6-95) Removed all OWL-related code, Added GUIConfig

1.001 (JohnV 02-29-96) Added additional plots, cleaned up destructor.

=====================================================================
*/

// #ifdef UNIX
#include "diags.h"
// #endif

#include "plsmadev.h"
#include "sptlrgn.h"
#include "advisman.h"
#include "ptclgrp.h"
#include "spatialg.h"
#ifndef UNIX
// #include "scaliter.h"
// #include "guiconfg.h"
#endif

PlasmaDevice::PlasmaDevice(SpatialRegionList* spatialRegionList)
{
	J = 0;
	K = 0;
	time = 0;
	ptclCount=0;
	setSpatialRegionList(spatialRegionList);
	particle_limit = 100000000;  //High limit by default, 100 million
	guiPtclCount = new Scalar[(int)speciesList->nItems()];
}

PlasmaDevice::~PlasmaDevice()
{
	// will need to delete all physics objects here
//	deleteSpatialRegions();
	spatialRegionList->deleteAll();
	delete spatialRegionList;
	spatialRegionList = NULL;
	speciesList->deleteAll();
	delete speciesList;
	speciesList = NULL;
	delete[] guiPtclCount;
	guiPtclCount = NULL;
}



#ifndef UNIX
/**
void PlasmaDevice::setGUI(GUIConfig* gui)
{
	theGUI = gui;
	createPlotManagers();
}


void PlasmaDevice::createPlotManagers()
{
	if (theGUI) {
		theGUI->createPlotManager("Particle Plot", 2);
		theGUI->createPlotManager("Field Plot");
		theGUI->createPlotManager("History", 2);
		theGUI->createPlotManager("XY Plot");
		theGUI->createPlotManager("Slice Plot");
	}
}
*/
#endif

void PlasmaDevice::setSpatialRegionList(oopicList<SpatialRegion>* _spatialRegionList)
{
	spatialRegionList = _spatialRegionList;

	oopicListIter<SpatialRegion> srIter(*spatialRegionList);
	theSpace = srIter.current();

	J = theSpace->getJ();
	K = theSpace->getK();
	speciesList = theSpace->getSpeciesList();

	switch (theSpace->query_geometry())
	{
		case ZRGEOM:
			x1Label = "z [m]";
			x2Label = "r [m]";
			x3Label = "phi [rad]";
			break;
		case ZXGEOM:
			x1Label = "z [m]";
			x2Label = "x [m]";
			x3Label = "y [m]";
			break;
	}
}

void PlasmaDevice::setAdvisor(AdvisorManager* theAdvisor)
{
	advisor = theAdvisor;
}

void PlasmaDevice::advance()
{
//	ptclCount = ParticleCount();
//	if (ptclCount > particle_limit) {
//		ListIter<SpatialRegion> srIter(*spatialRegionList);
//		for (srIter.restart(); !srIter.Done();	srIter++)
//			srIter.current()->increaseParticleWeight();  // reduce the #particles
//	}	                                                // and double their weight
	oopicListIter<SpatialRegion> srIter(*spatialRegionList);
	for (srIter.restart(); !srIter.Done();	srIter++)
		srIter()->advance();
}

Scalar PlasmaDevice::AdvancePIC()					// return current time
{
	advance();
	time = theSpace->getTime();
	return time;
}

#ifndef UNIX
/*
int PlasmaDevice::getGridJ()
{
	return J;
}

int PlasmaDevice::getGridK()
{
	return K;
}


void PlasmaDevice::getCavSize(Scalar& _m1s, Scalar& _m1f,
	Scalar& _m2s, Scalar& _m2f)
{
	_m1s = theSpace->getMKS(0, 0).e1();
	_m2s = theSpace->getMKS(0, 0).e2();
	_m1f = theSpace->getMKS(J, K).e1();
	_m2f = theSpace->getMKS(J, K).e2();
}

void PlasmaDevice::DefinePlots()
{
	theGUI->clearPlotData();
	theGUI->setTime(&time);

	DefineParticlePlots();
	DefineXYPlots();
	DefineHistoryPlots();
	DefineSlicePlots();
	DefineFieldPlots();
//	DefineAdditionalPlots();

	createProbeHistories();

	theGUI->reconnectPlotData(J, K);
}
*/

/**
void PlasmaDevice::DefineXYPlots()
{
*/
/*
	PlotDefn* xyPlotDefn;

	// should contain zero or more calls like:

	//  	plotName = "Electron V1 vs z";
	//		xLabel = "z [m]", yLabel = "V1 [m/s]";
	//
	//>>> 	xyPlotDefn = xyPlotMgr->DefineXYPlot(plotName, xLabel, yLabel,
	//							xLinlog, yLinlog, xAutoScale, yAutoScale,
	//							dfltXmin, dfltXmax, dfltYmin, dfltYmax;
	//
	//  	quantityName1 = "electron";
	//		xAxisId = "z";
	//		yAxisId = "V1";
	//		XIter1 = new AllXYDataIter(xData1, &dataSize1);
	//		YIter1 = new AllXYDataIter(yData1, &dataSize1);
	//
	//>>>	xyPlotDefn->SetXYData(quantityName1, xAxisId, yAxisId, XIter1, YIter1);
	//
	//----------------------------------
	xyPlotDefn = xyPlotMgr->DefineXYPlot("Field at [3,4] vs Time", "time", "field", linear, linear,
									TRUE, TRUE, 0, 1000, -10, 10);

	HistoryRecorder* b3Recorder = new HistoryRecorder;
	localHistories.add( b3Recorder );
	b3Recorder->Reset( &time, 1.0e12, ((Scalar*)&(theSpace->getBNode())[3][4]) + 2, 1.0 );
	xyPlotDefn->SetXYData ("B3/time", "time", "B3[3,4]", b3Recorder->XIter, b3Recorder->YIter);

	HistoryRecorder* e1Recorder = new HistoryRecorder;
	localHistories.add( e1Recorder );
	e1Recorder->Reset( &time, 1.0e12, (Scalar*)&(theSpace->getENode())[3][4], 1.0 );
	xyPlotDefn->SetXYData ("E1/time", "time", "E1[3,4]", e1Recorder->XIter, e1Recorder->YIter);

	//----------------------------------
*/
/*
	xyPlotDefn = xyPlotMgr->DefineXYPlot("Slice of EField at X[2]", "X[2]", "eField[2]", linear, linear,
									TRUE, TRUE, 0, theSpace->getK(), -10000, 10000);

	AllXYDataIter* xIter = (AllXYDataIter*) new AllScalarDataIter( (theSpace->getX())[2], &K, x2);
	AllXYDataIter* yIter = (AllXYDataIter*) new AllVctrFldsIter( (theSpace->getENode())[2], &K, e1);
	xyPlotDefn->SetXYData ("E1[2]/X[2]", "X[2]", "E1[2]", xIter, yIter);
*/
/*

	//--------------------------------------
	xyPlotDefn = xyPlotMgr->DefineXYPlot("Electron Count", "time", "n", linear, linear,
								TRUE, TRUE, 0, 1000, 0, 10000);
	HistoryRecorder* ptclCountRecorder = new HistoryRecorder;
	localHistories.add( ptclCountRecorder );
	ptclCountRecorder->Reset( &time, 1.0e12, &time, 1.0 );

	xyPlotDefn->SetXYData ("n/time", "time", "n", ptclCountRecorder->XIter, ptclCountRecorder->YIter);

	//---------------------------------
	PlotQuickSelector* qs;
	qs = xyPlotMgr->DefinePlotQuickSelector( "XY Plots", 1, 0 );

	qs->SetLabel( LEFT, L1, "electron n/t", "History Plot of electron n vs t [pS]", "n", "time" );

	qs->SetPlot( LEFT, L1, "Electron Count", "n/time", "n", "n/time", "time" );		// ordinate, abscissa

	//---------------------------------
	// the following code is optional
	// it customizes the xy QuickSelector
*/
/*
}
*/


/*
void PlasmaDevice::DefineHistoryPlots()
{
	// create (or recreate) the plotdefn object for history plots
	theGUI->DefinePlot("History", "Particle Count", "t [ps]", "N",
		linear, linear, TRUE, TRUE, 0, 1000, 0, 1000);

	// create the history QuickSelector
	// this must be called before the QS items
	theGUI->createQuickSelector("History", "any name", theSpace->get_nSpecies(), 0);

	ListIter<Species> sIter(*speciesList);

	for (int i=0; i < theSpace->get_nSpecies(); i++, sIter++)
	{
		const char* str = sIter.current()->get_name().c_str();
		theGUI->SetHistoryPlotData("History", "Particle Count",
			str, "time", "n", &time, &guiPtclCount[sIter.current()->getID()],
			line, white-i);
		// now define the QS items
		theGUI->setQSItem("History", 0,  str, "Particle Count", str, "n",
			str, "time");
	}
}

void PlasmaDevice::DefineSlicePlots()
{
}

void PlasmaDevice::DefineParticlePlots()
{
	// create (or recreate) the plotdefn object for particle plots
	theGUI->DefinePlot("Particle Plot", "Particle", x1Label, x2Label,
		"electron", linear, linear, linear, FALSE, TRUE, TRUE,
		0, J, 0, K,
//		theSpace->getMKS(0, 0).e1(), theSpace->getMKS(0, 0).e2(),
//		theSpace->getMKS(J, K).e1(), theSpace->getMKS(J, K).e2(),
		0, 0);

	ParticleSpaceIter* X1Iter =
		new ParticleSpaceIter(spatialRegionList, ParticleGroup::get_xPtr, 1);
	ParticleSpaceIter* X2Iter =
		new ParticleSpaceIter(spatialRegionList, ParticleGroup::get_xPtr, 2);
	ParticleSpaceIter* V1Iter =
		new ParticleSpaceIter(spatialRegionList, ParticleGroup::get_uPtr, 1);
	ParticleSpaceIter* V2Iter =
		new ParticleSpaceIter(spatialRegionList, ParticleGroup::get_uPtr, 2);
	ParticleSpaceIter* V3Iter =
		new ParticleSpaceIter(spatialRegionList, ParticleGroup::get_uPtr, 3);

	theGUI->SetPlotData("Particle Plot", "Particle", "electron",
		"z", "r", "u1", "u2", "u3",
		X1Iter, X2Iter, V1Iter, V2Iter, V3Iter);

	theGUI->createQuickSelector("Particle Plot", "Particle Plots", 6, 0);
	theGUI->setQSItem("Particle Plot", 0,  "x2 vs x1", "Particle", "electron",
		"r",	"electron", "z");
	theGUI->setQSItem("Particle Plot", 1,  "u1 vs x1", "Particle", "electron", "u1",
		"electron", "z");
	theGUI->setQSItem("Particle Plot", 2,  "u2 vs x1", "Particle", "electron", "u2",
		"electron", "z");
	theGUI->setQSItem("Particle Plot", 3,  "u3 vs x1", "Particle", "electron", "u3",
		"electron", "z");
	theGUI->setQSItem("Particle Plot", 4,  "u1 vs x2", "Particle", "electron", "u1",
		"electron", "r");
	theGUI->setQSItem("Particle Plot", 5,  "u2 vs x2", "Particle", "electron", "u2",
		"electron", "r");
}

void PlasmaDevice::DefineFieldPlots()
{
	theGUI->DefinePlot("Field Plot", "Field", x1Label, x2Label, "E [V/m]",
		linear, linear, linear, TRUE, TRUE, TRUE,
		0, J/1000.0, 0, K/1000.0, 0, 100);

	ScalarIter* x1Iter =  new FieldScalarIter(theSpace->getX(),
		theSpace->getJ()+1, theSpace->getK()+1, 1);
	ScalarIter* x2Iter =  new FieldScalarIter(theSpace->getX(),
		theSpace->getJ()+1, theSpace->getK()+1, 2);

	ScalarIter* E1Iter =  new FieldScalarIter(theSpace->getENode(),
		theSpace->getJ()+1, theSpace->getK()+1, 1);
	ScalarIter* E2Iter =  new FieldScalarIter(theSpace->getENode(),
		theSpace->getJ()+1, theSpace->getK()+1, 2);
	ScalarIter* E3Iter =  new FieldScalarIter(theSpace->getENode(),
		theSpace->getJ()+1, theSpace->getK()+1, 3);

	theGUI->SetPlotData("Field Plot", "Field", "efield", x1Label, x2Label, "E1", "E2", "E3",
		x1Iter, x2Iter, E1Iter, E2Iter, E3Iter, mesh, ltBlue);

*/
/*	theGUI->DefinePlot("Field Plot", "B Field", "z [m]", "r [m]", "B [ T ]",
		linear, linear, linear, TRUE, TRUE, TRUE,
		0, J/1000.0, 0, K/1000.0, 0, 100);   */

// It seems these are redundant (JohnV 03-08-96):
//	ScalarIter* bX1Iter =  new FieldScalarIter(theSpace->getX(),
//		theSpace->getJ()+1, theSpace->getK()+1, 1);
//	ScalarIter* bX2Iter =  new FieldScalarIter(theSpace->getX(),
//		theSpace->getJ()+1, theSpace->getK()+1, 2);

/*
	ScalarIter* b1Iter =  new FieldScalarIter(theSpace->getBNode(),
		theSpace->getJ()+1, theSpace->getK()+1, 1);
	ScalarIter* b2Iter =  new FieldScalarIter(theSpace->getBNode(),
		theSpace->getJ()+1, theSpace->getK()+1, 2);
	ScalarIter* b3Iter =  new FieldScalarIter(theSpace->getBNode(),
		theSpace->getJ()+1, theSpace->getK()+1, 3);

	theGUI->SetPlotData("Field Plot", "Field", "bfield", x1Label, x2Label, "B1", "B2", "B3",
		x1Iter, x2Iter, b1Iter, b2Iter, b3Iter, mesh, red);

*/
/*
	ScalarIter* i1Iter =  new FieldScalarIter(theSpace->getI(),
		theSpace->getJ()+1, theSpace->getK()+1, 1);
	ScalarIter* i2Iter =  new FieldScalarIter(theSpace->getI(),
		theSpace->getJ()+1, theSpace->getK()+1, 2);
	ScalarIter* i3Iter =  new FieldScalarIter(theSpace->getI(),
		theSpace->getJ()+1, theSpace->getK()+1, 3);
	theGUI->SetPlotData("Field Plot", "Field", "ifield", "z", "r", "I1", "I2", "I3",
		x1Iter, x2Iter, i1Iter, i2Iter, i3Iter, mesh, cyan);

	int count = theSpace->getDiagList()->nItems();
*/

	// theGUI->createQuickSelector("Field Plot", "Field Plots", 6 /*+count*/, 0);
	// seems to be hardwired for a max of 6 items!!!!
/*
	theGUI->setQSItem("Field Plot", 0,  "E1", "Field",
		"efield", "E1", "efield", x1Label, "efield", x2Label);
	theGUI->setQSItem("Field Plot", 1,  "E2", "Field",
		"efield", "E2", "efield", x1Label, "efield", x2Label);
	theGUI->setQSItem("Field Plot", 2,  "E3", "Field",
		"efield", "E3", "efield", x1Label, "efield", x2Label);
	theGUI->setQSItem("Field Plot", 3,  "B1", "Field",
		"bfield", "B1", "bfield", x1Label, "bfield", x2Label);
	theGUI->setQSItem("Field Plot", 4,  "B2", "Field",
		"bfield", "B2", "bfield", x1Label, "bfield", x2Label);
	theGUI->setQSItem("Field Plot", 5,  "B3", "Field",
		"bfield", "B3", "bfield", x1Label, "bfield", x2Label);

	ListIter<Diag> diagIter(*theSpace->getDiagList());
	for (int i=5; !diagIter.Done(); diagIter++, i++)
		diagIter.current()->initwin(theGUI, i);
}

void PlasmaDevice::DefineAdditionalPlots()
{
}

long PlasmaDevice::ParticleCount()
{
	if (theSpace) return theSpace->nParticles(guiPtclCount);
	else return 0;
}
*/

#endif

Scalar PlasmaDevice::SimulationTime()
{
	return time;
}

// called from PlasmaDevice
//
#ifndef UNIX
/*
void PlasmaDevice::createProbeHistories()
{
	oopicList<SpatialRegionGroup>* SRGL = advisor->getSpatialRegionGroupList();
	oopicListIter<SpatialRegionGroup> srgi(*SRGL);
	oopicListIter<SpatialRegion> sri(*spatialRegionList);
	for (srgi.restart(), sri.restart();
			!(srgi.Done() || srgi.Done()); srgi++, sri++) {
		srgi.current()->createProbeHistories(theGUI, sri.current());
	}
}

*/
#endif

int PlasmaDevice::Dump(FILE* DMPFile)
{
	SpatialRegionList reversedRegionList;
	
	oopicListIter <SpatialRegion> Siter(*spatialRegionList);
  for(Siter.restart();!Siter.Done();Siter++) 
		reversedRegionList.add(Siter());

  oopicListIter <SpatialRegion> Riter(reversedRegionList);
  for(Riter.restart();!Riter.Done();Riter++) 
		Riter()->Dump(DMPFile);

	return (1);

}

