/*
====================================================================

PORTGAUSS.CPP

====================================================================
*/

#include "port.h"
#include "portgauss.h"
#include "fields.h"
#include "ptclgrp.h"

PortGauss::PortGauss(oopicList <LineSegment> *segments, int _EFFlag,
           Scalar _gauss_w, Scalar _gauss_rl, Scalar _omegag,
           Scalar _zf, Scalar _Agw) : Boundary(segments)
{

 if(segments->nItems() > 1)
    {
    cout << "Warning, Port-derived boundaries can only have 1 segment.\n";
    cout << "Check your input file.\n";
    }

 BCType=DIELECTRIC_BOUNDARY;
 EFFlag = _EFFlag;
 gauss_w = _gauss_w;
 gauss_rl = _gauss_rl;
 omegag = _omegag;
 zf = _zf;
 Agw = _Agw;

 dt = 0;
 tOld = 0;
 
 if(EFFlag)
    {
    EnergyFlux = &EnergyOut;
    EnergyOut =0.0;

    if (normal==1)
       shift = 0;
    else
    shift = -1;
    }

}

//--------------------------------------------------------------------
//	Apply boundary conditions to fields locally.


void PortGauss::applyFields(Scalar t, Scalar dt)
{

 time = t- dt/2;
 grid = fields->get_grid();

 if (fields->getSub_Cycle_Iter() ==1) EnergyOut = 0.0;
 else if (fields->getSub_Cycle_Iter() ==0) EnergyOut = 0.0;

 A1 =  get_time_value((float)t);

 zb = grid->getMKS(j1,k1).e2();
 zp02 = gauss_rl*gauss_rl;
 zp2 = (zb - zf)*(zb - zf);
 wz = gauss_w*sqrt(1. + zp2/zp02);
 wz2 = wz*wz;
 Rz = (zb - zf) + zp02/(zb - zf);
 phig = atan((zb-zf)/gauss_rl);
 kg = 2.*gauss_rl/(gauss_w*gauss_w);

 if (grid->query_geometry() == ZRGEOM)
    {
    cout << "Gaussian wave is not implemented for cylindrical geometry"
          << endl;
    }

 if (grid->query_geometry() == ZXGEOM)
    {
    if (alongx2())
       {
       a = (grid->getMKS(j2,k2) - grid->getMKS(j1,k1)).e2();

       for (int k = k1; k < k2; k++)
          {
	  Scalar yforEy = (grid->getMKS(j1, k + 1).e2() +
	  grid->getMKS(j1, k).e2())/2.0 - grid->getMKS(j1,k1).e2() ;

          Scalar rt2 = (yforEy - 0.5*a)*(yforEy - 0.5*a);
          Scalar Ey = (A1*Agw*gauss_w*exp(-rt2/wz2)/wz)*(
                      cos(0.5*kg*rt2/Rz + kg*(zb) - phig)*cos(omegag*t) +
                      sin(0.5*kg*rt2/Rz + kg*(zb) - phig)*sin(omegag*t));


	  Scalar intEy = grid->dl2(j1, k)*Ey;
	  fields->setIntEdl(j1, k, 2, intEy);

	  if(EFFlag)
	     {
	     if (fields->getiC()[j2+normal][k].e2())
	       {
	       HonBoundary = (fields->getiL()[j2+shift][k].e3())*
	       (fields->getIntBdS()[j2+shift][k].e3()) + 
	       normal*.5*(fields->getIMesh(j2,k).e2() + 
	       get_time_value_deriv((float)time)/(fields->getiC()[j2+normal][k].e2()));
	       EnergyOut += normal*Ey*HonBoundary*grid->dS(j2+shift,k).e1();
	       }
	     }
          }
       }
    }

 if (fields->getFieldSub() == fields->getSub_Cycle_Iter()) EnergyOut
      /= (fields->getFieldSub());
}

//--------------------------------------------------------------------
//	Set the passive bc for fields at the port.  Currently just a
//	copy of those for conductor.


void PortGauss::setPassives()
{

 if (alongx2())					//	vertical
    {

    for (int k=MIN(k1,k2); k<MAX(k1,k2); k++)
       {
       fields->set_iC2(j1, k, 0);       	
       fields->set_iC3(j1, k, 0);
       }

    fields->set_iC3(j1, k2, 0);
    }
}

//--------------------------------------------------------------------
//	Port::emit() simply deletes Particles in its stack.  May add some
//	diagnostics for particles collected in the future.

#ifndef __linux__
#pragma argsused
#endif

ParticleList& PortGauss::emit(Scalar t, Scalar dt, Species *species)
{
 while(!particleList.isEmpty())
    {
    Particle* p = particleList.pop();
    delete p;
    }

 return particleList;
}

