/*
=======================================================================

secondary.cpp

These classes represent a number of secondary models:

Secondary: simple step function secondary emission, with turn-on threshold.

Secondary2: a Vaughan-based model, includes energy and angular dependence, 
full emission spectrum including reflected and scattered primaries.

1.0 (JohnV, 09-02-98) Created by cutting Secondary from dielectric.h, added
    Secondary2.

=======================================================================
*/

#include "secondary.h"
#include "boundary.h"
#include "vmaxwelln.h"

Secondary::Secondary(Scalar _secondary, Scalar _threshold, Scalar Eemit, 
							Species* _secSpecies, Species* _iSpecies)
{
  secondary = _secondary; 
  threshold = _threshold; 
  secSpecies = _secSpecies; 
  iSpecies = _iSpecies;
  v_emit = sqrt(2*1.602e-19*Eemit/secSpecies->get_m()); 
printf("Secondary\n");
}

/*
Secondary::~Secondary()
{

}
*/

void Secondary::setBoundaryPtr(Boundary* _bPtr)
{
  bPtr = _bPtr;
  if (bPtr->alongx1()) normal.set_e2(1);
  else normal.set_e1(1);
}

int Secondary::generateSecondaries(Particle& p, ParticleList& pList)
{
  if (p.get_speciesPtr() != iSpecies) return 0;
  if (p.energy_eV_MKS() < threshold) return 0;
  Scalar yield = number(p);
  int n = (int)yield;
  yield -= n;
  if (frand() <= yield) n++;
  for (int i=0; i<n; i++) pList.add(createSecondary(p));
  return n;
}

// number of secondaries emitted per incident particle

Scalar Secondary::number(Particle& p)
{
  return secondary;
}

Particle* Secondary::createSecondary(Particle& p)
{
  return new Particle(p.get_x(), frand()*v_emit*bPtr->unitNormal(), secSpecies, 
							 p.get_np2c(), p.isVariableWeight());
}

Secondary2::Secondary2(Scalar secondary, Scalar threshold, Scalar Eemit, 
							  Species* secSpecies, Species* iSpecies, Scalar f_ref,
							  Scalar f_scat, Scalar _energy_w, Scalar _energy_max0,
							  Scalar _ks)
: Secondary(secondary, threshold, Eemit, secSpecies, iSpecies)
{
  fractionReflected = f_ref;
  fractionScattered = f_scat;
  energy_w = Eemit;
  energy_max0 = _energy_max0;
  ks = _ks;
  // dist = new MaxwellianFlux(Eemit, secSpecies, EV);
  dist = new MaxwellianFlux(secSpecies);
  dist->setThermal(Eemit, EV);
printf("Secondary2\n");
}

Secondary2::~Secondary2()
{
  delete dist;
}

Scalar Secondary2::number(Particle& p)
{
  if (secondary <= 0) return 0;
  Scalar theta;
  Vector3 v = p.get_u()/p.gamma();
  theta = acos(-(v*bPtr->unitNormal())/v.magnitude());
  Scalar Emax = energy_max0*(1 + ks*theta*theta/M_PI);
  Scalar w = (p.energy_eV_MKS() - threshold)/(Emax - threshold);
  Scalar k = (w > 1) ? 0.25 : 0.62;
  return secondary*(1 + 0.5*ks*theta*theta/M_PI)*pow(w*exp(1-w), k); 
}

Particle* Secondary2::createSecondary(Particle& p)
{
  Scalar R = frand();
  Vector3 u;
  if (R < fractionScattered+fractionReflected){
	 u = p.get_u();
	 if (bPtr->alongx1()) u.set_e2(-u.e2()); // reflected primary;
	 else u.set_e1(-u.e1());
	 if (R > fractionReflected) u *= frand(); // scattered primary;
  }
  else u = dist->get_U(bPtr->unitNormal()); // true secondary;
//printf("u=(%g,%g,%g)\n", u.e1(), u.e2(), u.e3());
  return new Particle(p.get_x(), u, secSpecies, p.get_np2c(), 
							 p.isVariableWeight());
}
