/*
====================================================================

spblink.cpp

Purpose:        Defines implementation for SPBoundLink, which sends and
                receives particles and fields through a boundary

Version:        $Id: spblink.cpp,v 1.14.2.15 2000/03/07 13:41:24 cary Exp $

The communication link between pairs of spatial region boundary

0.99    (PeterM 9-3-97) Inception
====================================================================
*/

#include "spblink.h"
#ifdef MPI_VERSION
extern int MPI_RANK;  //globally defined MPI_RANK
#endif

// To get the precision of MPI transfers
#include <precision.h>

SPBoundLink::SPBoundLink(Vector3 *_lE, Vector3 *_lB, Vector3 *_lJ,
	int _lenIn,int _lenDest) {
  lE=_lE;
  lB=_lB;
  lJ=_lJ;
  lenIn = _lenIn;
  lenDest = _lenDest;
  LinkType = SAME_PROCESS;  //  default, may be overridden by makeLink()
  PinArraySize=100;
  PinArray = new particlePassDat[PinArraySize];

// Buffers for shifted particles
  shftPtclsSize = 100;
  shftPtcls = new ShiftedParticlePassData[shftPtclsSize];
  numShftPtcls = 0;
#ifdef MPI_VERSION
  shftPtclSendRequest = MPI_REQUEST_NULL;
#endif

// Buffers for usual fields
  fieldsBufout = new Vector3[3 * _lenIn + 6];
  fieldsBufin = new Vector3[3 * _lenIn + 6];
  int i;
  for(i=0; i < 3 * _lenIn +6; i++) 
	fieldsBufin[i]=fieldsBufout[i]=Vector3(0,0,0);

// Buffers for shifted fields
#ifdef MPI_VERSION
  shftFldSendRequest = MPI_REQUEST_NULL;
  shftFldRecvRequest = MPI_REQUEST_NULL;
#endif
  shiftedFieldsSent = false;
  shftFldsBufout = new Vector3[3 * _lenIn + 6];
  shftFldsBufin = new Vector3[3 * _lenIn + 6];
  for(i=0; i < 3 * _lenIn +6; i++) 
	shftFldsBufin[i]=shftFldsBufout[i]=Vector3(0,0,0);

}

// Destructor - return memory
//
SPBoundLink::~SPBoundLink(){
  delete [] PinArray;
  delete [] shftPtcls;
  delete [] fieldsBufout;
  delete [] fieldsBufin;
  delete [] shftFldsBufout;
  delete [] shftFldsBufin;
}

void SPBoundLink::makeLink(Vector3 *_lEdest, Vector3 *_lBdest,
	Vector3 *_lJdest, Vector3 **_lE, Vector3 **_lB, Vector3 **_lJ)
{
  LinkType = SAME_PROCESS;
  
  *_lE = lE;  // set the other guy's pointer to point to our receiving area
  *_lB = lB;  // set the other guy's pointer to point to our receiving area
  *_lJ = lJ;  // set the other guy's pointer to point to our receiving area

  lEdest = _lEdest; // set our pointers to point at the remote receiving area
  lBdest = _lBdest; // set our pointers to point at the remote receiving area
  lJdest = _lJdest; // set our pointers to point at the remote receiving area

}

void SPBoundLink::makeLink(int _DestID,int _tag_modifier)
{
  LinkType = MPI_LINK;
  DestID = _DestID;  // tells us which boundary we're communicating with
  tag_modifier = _tag_modifier;  //

#ifdef MPI_DEBUG
  printf("\nSPBoundLink::makeLink: (%d)  DestID = %d, tag_modifier = %d",MPI_RANK,DestID,tag_modifier);
#endif

}

void SPBoundLink::sendParticles(int n, particlePassDat *passers) {
  switch(LinkType) {

    case SAME_PROCESS:
      break;

#ifdef MPI_VERSION
    case MPI_LINK:
      {
#ifdef MPI_DEBUG
        printf("\nSPBoundLink::sendParticles (%d):  DestID = %d, tag = %d\n",
		MPI_RANK,DestID,P_TAG+tag_modifier);
#endif

//#define DEBUG
#ifdef DEBUG
        if(n > 0)
        printf("\nSBL:passParticles:  sending %d particles to %d",n,DestID);
#endif
        MPI_Isend(passers, (sizeof(particlePassDat)/sizeof(Scalar))*n,
		MPI_Scalar,DestID, P_TAG + tag_modifier,XOOPIC_COMM,&Psrequest);
      }
      break;

#endif
  }
}

void SPBoundLink::askParticles() {

  switch(LinkType) {

    case SAME_PROCESS:
      break;

#ifdef MPI_VERSION
    case MPI_LINK:
      {
// non-blocking receive NOT possible
      }
      break;
#endif

  }

}

particlePassDat * SPBoundLink::waitParticles(int *n_in) {

  switch(LinkType) {
    case SAME_PROCESS:
      break;

#ifdef MPI_VERSION
    case MPI_LINK:
      {
	int N_in;
  #ifdef MPI_DEBUG
        printf("\nSPBoundLink::waitParticles (E %d):  DestID = %d, tag = %d\n",
		MPI_RANK,DestID,P_TAG+tag_modifier);
  #endif

// wait for our send to complete
	MPI_Wait(&Psrequest,&Psstatus);

// count how many particles are incoming
	MPI_Probe(DestID,P_TAG+tag_modifier,XOOPIC_COMM,&Prstatus);
	MPI_Get_count(&Prstatus,MPI_Scalar,&N_in); 

// if our buffer isn't large enough, increase the size
	while((*n_in=N_in*sizeof(Scalar)/sizeof(particlePassDat)) > PinArraySize)
		{
	  particlePassDat *array = new particlePassDat[2*PinArraySize];
	  PinArraySize*=2;
	  delete [] PinArray;
	  PinArray=array;
	}

// receive the incoming particles
        MPI_Recv(PinArray, N_in, MPI_Scalar, DestID, P_TAG+tag_modifier,
		XOOPIC_COMM, &Prstatus);

  #ifdef DEBUG
	printf("\nSBL:waitParticles:  getting %d particles from %d", N_in,
		DestID);
  #endif

        return PinArray;
      }
      break;

#endif
  }

  return PinArray;

}

void SPBoundLink::sendFields(Vector3 *_lEsend, Vector3 *_lBsend,
	Vector3 *_lJsend) {

  switch(LinkType) {

    case SAME_PROCESS:
      {
// Copy the data to be sent over to the other boundary
        for(int i=0;i<lenIn;i++) {
	  lEdest[i] = _lEsend[i];
	  lBdest[i] = _lBsend[i];
	  lJdest[i] = _lJsend[i];
// I think the following are not needed
#ifdef MPI_VERSION
	  // fieldsBufout[i] = _lEsend[i];
	  // fieldsBufout[i+lenIn+2] = _lBsend[i];
	  // fieldsBufout[i+2*lenIn+4] = _lJsend[i];
#endif //MPI_VERSION		
        }
      }
      break; 

#ifdef MPI_VERSION
    case MPI_LINK:
      {
        for(int i=0;i<lenIn;i++) {
	  fieldsBufout[i] = _lEsend[i];
	  fieldsBufout[i+lenIn+2] = _lBsend[i];
	  fieldsBufout[i+2*lenIn+4] = _lJsend[i];
	}
// send one message instead of 3
	MPI_Isend(fieldsBufout, 9 * lenIn+18, MPI_Scalar,
		DestID, E_TAG + tag_modifier, XOOPIC_COMM, &Esrequest);
      }
      break;
#endif

  }

}

void SPBoundLink::askFields() {

  switch(LinkType) {

    case SAME_PROCESS:
      {
      }
      break;  // nothing needs doing. 

#ifdef MPI_VERSION
    case MPI_LINK:
      {
	MPI_Irecv(fieldsBufin, 9*lenIn+18, MPI_Scalar,
		DestID, E_TAG + tag_modifier, XOOPIC_COMM, &Errequest);
      }
      break;
#endif

  }

}

void SPBoundLink::waitFields() {

  switch(LinkType) {

    case SAME_PROCESS:
      {
// nothing needs doing. 
      }
      break;

#ifdef MPI_VERSION
    case MPI_LINK:
      {

//  make sure all the sends finish
	MPI_Wait(&Esrequest,&Esstatus);

//  make sure all the receives finish
	MPI_Wait(&Errequest,&Erstatus);

// copy the fields back out of the receive buffer.
	for(int i=0;i<lenIn;i++) {
	  lE[i]=fieldsBufin[i];
	  lB[i]=fieldsBufin[i+lenIn+2];
	  lJ[i]=fieldsBufin[i+2*lenIn+4];
	}
      }
      break;
#endif

  }

} 

void SPBoundLink::sendShiftedParticles(vector<Particle*>& pl){

// Make sure local buffer is at least as large as the particle list
  if(shftPtclsSize < pl.size()){
    delete [] shftPtcls;
    shftPtcls = new ShiftedParticlePassData[pl.size()];
    shftPtclsSize = pl.size();
  }

// Copy the particles into the list
  for(int i=0; i<pl.size(); i++){
    shftPtcls[i].speciesID = (Scalar) pl[i]->get_speciesID();
    shftPtcls[i].x = pl[i]->get_x();
    shftPtcls[i].u = pl[i]->get_u();
    shftPtcls[i].np2c = pl[i]->get_np2c();
    if ( pl[i]->isVariableWeight() ) shftPtcls[i].vary_np2c = 1.;
    else shftPtcls[i].vary_np2c = 0.;
  }
  numShftPtcls = pl.size();

// Send the particle over
  switch(LinkType) {

    case SAME_PROCESS:
      cout << "sendShiftedParticles not implemented for same process.\n";
      assert(0);
      break;

#ifdef MPI_VERSION
    case MPI_LINK:
      {

#ifdef DEBUG
        // cout << "\nSBL:passParticles:  sending " << n << " particles to " <<
		// DestID << endl;
#endif
        MPI_Isend(shftPtcls, 
	  (sizeof(ShiftedParticlePassData)/sizeof(Scalar))*numShftPtcls,
		MPI_Scalar, DestID, SHIFT_PTCL_TAG + tag_modifier, XOOPIC_COMM,
		&shftPtclSendRequest);
  	// cout << "MPI_RANK = " << MPI_RANK << 
	// ": SPBoundLink::sendShiftedParticles: shftPtclSendRequest = " <<
		// shftPtclSendRequest << endl;
      }
      break;
#endif

  }

}

// Wait for the particles from the boundary.
// Return reference to list with particles.

ParticleList& SPBoundLink::recvShiftedParticles(Species** speciesArray){

// Empty out old particles
  // cout << "MPI_RANK = " << MPI_RANK <<
        // ": deleting old particles in SPBoundLink::recvShiftedParticles." << endl;
  shftPtclList.deleteAll();
  // cout << "MPI_RANK = " << MPI_RANK <<
        // ": old particles deleted in SPBoundLink::recvShiftedParticles." << endl;

// Get new particles from boundary
  switch(LinkType) {

    case SAME_PROCESS:
      cout << "waitShiftedParticles not implemented for same process.\n";
      assert(0);
      break;

#ifdef MPI_VERSION
    case MPI_LINK:
      {

// Wait for our send to complete
  	// cout << "MPI_RANK = " << MPI_RANK << ": waiting for receive " <<
        	// "to complete in SPBoundLink::recvShiftedParticles." << endl;
  	// cout << "MPI_RANK = " << MPI_RANK << ": shftPtclSendRequest = " <<
		// shftPtclSendRequest << endl;
	if(shftPtclSendRequest != MPI_REQUEST_NULL)
		MPI_Wait(&shftPtclSendRequest, &shftPtclSendStatus);

// Count how many particles are incoming
  	// cout << "MPI_RANK = " << MPI_RANK << ": probing " <<
        	// "in SPBoundLink::recvShiftedParticles." << endl;
	MPI_Probe(DestID, SHIFT_PTCL_TAG + tag_modifier, XOOPIC_COMM,
		&shftPtclRecvStatus);
        int numScalarIn;
  	// cout << "MPI_RANK = " << MPI_RANK << ": getting count " <<
        	// "in SPBoundLink::recvShiftedParticles." << endl;
	MPI_Get_count(&shftPtclRecvStatus, MPI_Scalar, &numScalarIn); 
        int numPtclsIn = numScalarIn * sizeof(Scalar) /
		sizeof(ShiftedParticlePassData);
  	// cout << "MPI_RANK = " << MPI_RANK << ": receiving " << numPtclsIn <<
        	// " particles in SPBoundLink::recvShiftedParticles." << endl;

// Local buffer must be at least as large as the number of incoming particles
        if(shftPtclsSize < numPtclsIn){
          delete [] shftPtcls;
          shftPtcls = new ShiftedParticlePassData[numPtclsIn];
          shftPtclsSize = numPtclsIn;
        }

// Receive the incoming particles
        MPI_Recv(shftPtcls, numScalarIn, MPI_Scalar, DestID, 
		SHIFT_PTCL_TAG + tag_modifier, XOOPIC_COMM, 
		&shftPtclRecvStatus);

// Add the particles to the list
        BOOL _vary_np2c;
        for(int i=0; i<numPtclsIn; i++) {
          if ( shftPtcls[i].vary_np2c > 0. ) _vary_np2c = TRUE;
          else _vary_np2c = FALSE;
	  Particle* pp = new Particle(shftPtcls[i].x, shftPtcls[i].u,
		speciesArray[(int)(shftPtcls[i].speciesID+0.1)],
		shftPtcls[i].np2c, _vary_np2c);
          shftPtclList.add(pp);
        }

      }
      break;
#endif	// MPI_VERSION

  }
  return shftPtclList;
}

void SPBoundLink::sendShiftedFields(Vector3 *_lEsend, Vector3 *_lBsend,
	Vector3 *_lJsend) {

  switch(LinkType) {

    case SAME_PROCESS:
      {
// Copy the data to be sent over to the other boundary
        for(int i=0;i<lenIn;i++) {
	  lEdest[i] = _lEsend[i];
	  lBdest[i] = _lBsend[i];
	  lJdest[i] = _lJsend[i];
// I think the following are not needed
#ifdef MPI_VERSION
	  // fieldsBufout[i] = _lEsend[i];
	  // fieldsBufout[i+lenIn+2] = _lBsend[i];
	  // fieldsBufout[i+2*lenIn+4] = _lJsend[i];
#endif //MPI_VERSION		
        }
      }
      break; 

#ifdef MPI_VERSION
    case MPI_LINK:
      {
        for(int i=0;i<lenIn;i++) {
	  shftFldsBufout[i] = _lEsend[i];
	  shftFldsBufout[i+lenIn+2] = _lBsend[i];
	  shftFldsBufout[i+2*lenIn+4] = _lJsend[i];
	}
// send one message instead of 3
	MPI_Isend(shftFldsBufout, 9 * lenIn+18, MPI_Scalar,
		DestID, SHIFT_FLD_TAG + tag_modifier, XOOPIC_COMM, 
		&shftFldSendRequest);
        shiftedFieldsSent = true;
      }
      break;
#endif

  }

}

void SPBoundLink::askShiftedFields() {

  switch(LinkType) {

    case SAME_PROCESS:
      {
      }
      break;  // nothing needs doing. 

#ifdef MPI_VERSION
    case MPI_LINK:
      {
	MPI_Irecv(shftFldsBufin, 9*lenIn+18, MPI_Scalar,
		DestID, SHIFT_FLD_TAG + tag_modifier, XOOPIC_COMM, 
		&shftFldRecvRequest);
        extern int MPI_RANK;
        // cout << "Request to receive fields made by MPI_RANK = " << 
		// MPI_RANK << endl;
      }
      break;
#endif

  }

}

void SPBoundLink::waitShiftedFields() {

  switch(LinkType) {

    case SAME_PROCESS:
      {
// nothing needs doing. 
      }
      break;

#ifdef MPI_VERSION
    case MPI_LINK:
      {

//  make sure all the sends finish
        if(shiftedFieldsSent) {
	  MPI_Wait(&shftFldSendRequest, &shftFldSendStatus);
	  shiftedFieldsSent = false;
	}

//  make sure all the receives finish
	MPI_Wait(&shftFldRecvRequest, &shftFldRecvStatus);

// copy the fields back out of the receive buffer.
	for(int i=0;i<lenIn;i++) {
	  lE[i]=shftFldsBufin[i];
	  lB[i]=shftFldsBufin[i+lenIn+2];
	  lJ[i]=shftFldsBufin[i+2*lenIn+4];
	}

/*
    int kout = 5;
    cout << "In SPBoundLink::waitShiftedFields:\n";
    cout << "E1 =";
    for(int j1=0; j1<=lenIn; j1++)  cerr << " " << shftFldsBufin[j1].e1();
    cout << endl;
    cout << "E2 =";
    for(int j1=0; j1<=lenIn; j1++)  cerr << " " << shftFldsBufin[j1].e2();
    cout << endl;
    cout << "I1 =";
    for(int j1=0; j1<=lenIn; j1++)  cout << " " << 
	shftFldsBufin[j1+2*lenIn+4].e1();
    cout << endl;
*/

      }
      break;
#endif

  }

} 
