//--------------------------------------------------------------------
//
// File:	spbound.cpp
//
// Purpose:	Implementation of the SpatialRegionBoundary - a boundary
//		between two spatial regions.  Has methods for passing
//		data between regions.
//
// Version:	$Id: spbound.cpp,v 1.27.2.18 2000/03/07 02:03:56 cary Exp $
//
//--------------------------------------------------------------------

/*
====================================================================
Revision History

Cary (23 Jan 00) Added header, implementation of recvParticles, which
	is like emit, except that one does not further weighting of
	the particles.  Did some indentation for readability.
CVS1.27.2.4 (Cary 26 Jan 00) Reorganized to put constructor first.
	Not necessary, I see now, but made sure that j indexing
	was increasing.  Implemented sendStripe and recvStripe.
	Still not checked as we have no way of looking at MPI runs.

====================================================================
*/

#include "spbound.h"
#include "fields.h"

SpatialRegionBoundary::SpatialRegionBoundary(oopicList <LineSegment> *segments) :
  Boundary(segments) {

	 if(segments->nItems() > 1) {
		cout << "Spatial region boundaries CANNOT have multiple segments.\n";
		exit(1);
	 }
	 
	 if(j1!=j2 && k1!=k2) {
		cout << "Spatial region boundaries MUST be either horizontal or vertical.\n";
		exit(1);
	 }

	 int length = MAX(abs(k1-k2),abs(j1-j2));

	 // memory allocation is 1 larger to allow guard cells for a
	 // simpler field solve.
	 lJ = new Vector3[length+1];
	 lE = new Vector3[length+1];
	 lB = new Vector3[length+1];
	 liL = new Vector3[length+1];
	 lJsend = new Vector3[length+1];
	 lEsend = new Vector3[length+1];
	 lBsend = new Vector3[length+1];

	 //zero the memory
	 memset(lJ,0,(length+1) * sizeof(Vector3));
	 memset(lE,0,(length+1) * sizeof(Vector3));
	 memset(lB,0,(length+1) * sizeof(Vector3));
	 memset(liL,0,(length+1) * sizeof(Vector3));

	 memset(lJsend,0,(length+1) * sizeof(Vector3));
	 memset(lEsend,0,(length+1) * sizeof(Vector3));
	 memset(lBsend,0,(length+1) * sizeof(Vector3));
	 theLink = 0;
	 BCType=SPATIAL_REGION_BOUNDARY;
	 ptclPassArraySize=100;
	 ptclPassArray = new particlePassDat[ptclPassArraySize];
	 number_Particles_passing=0;
	 

}

// We need to set up a special iC for the Fields:: so it doesn't interfere
// with the special field solve we do for E2 and E3 on the boundary.

void SpatialRegionBoundary::setPassives() {
  int j,k;
  if(j1==j2) // vertical boundary, set iC2 and iC3 to zero
	 {
		for(k=k1;k<=k2;k++) {
		  fields->set_iL1(j1,k,2*fields->getiL()[j1][k].e1());  // fix iL: default is wrong.
		  fields->set_iC2(j1,k,0);
		  fields->set_iC3(j1,k,0);
		}
	 }
  else  // horizontal boundary, set iC1 and iC3 to zero
	 {
		for(j=j1;j<=j2;j++) {
		  fields->set_iL2(j,k1,2*fields->getiL()[j][k1].e2());  // fix iL: default is wrong.
		  fields->set_iC1(j,k1,0);
		  fields->set_iC3(j,k1,0);
		}
	 }
}
		  

//  This is the function which will assemble and send the currents
//  and fields necessary for the other boundary to solve for its
//  fields.  By convention for this boundary, normal points IN

void SpatialRegionBoundary::putCharge_and_Current(Scalar t) {

  Vector3 **intBdS = fields->getIntBdS();
  Vector3 **intEdl = fields->getIntEdl();
  Vector3 **I = fields->getI();
  // Vector3 **iC = fields->getiC();
  // Vector3 **iL = fields->getiL();

// For indexing
  int kl = MIN(k1,k2);
  int kh = MAX(k1,k2);
  int jl = MIN(j1,j2);
  int jh = MAX(j1,j2);

// Begin the request
  if(!theLink) return;
  theLink->askFields();

// Horizontal boundary data send
  if(k1==k2) for(int j=jl;j<jh;j++) {  

    if(normal==-1) {
// Real region below
      lBsend[j-jl]=Vector3(intBdS[j][k1-1].e1(),0,intBdS[j][k1-1].e3());
      lEsend[j-jl]=intEdl[j][k1-1];
      lJsend[j-jl]=Vector3(I[j][k1].e1(),I[j][k1-1].e2(),I[j][k1].e3());
    }

    else {
// Real region above
      lBsend[j-jl]=Vector3(intBdS[j][k1].e1(), 0, intBdS[j][k1].e3());
      lEsend[j-jl]=Vector3(intEdl[j][k1].e1(), intEdl[j][k1+1].e2(),
		intEdl[j][k1+1].e3());
      lJsend[j-jl]=Vector3(I[j][k1].e1(), I[j][k1].e2(), I[j][k1].e3());
    }

  }

// Vertical boundary data send
  else for(int k=kl;k<kh;k++) {  

// Real region left
    if(normal==-1) {
      lBsend[k-kl]=Vector3(0,intBdS[j1-1][k].e2(),intBdS[j1-1][k].e3());
      lEsend[k-kl]=intEdl[j1-1][k];
      lJsend[k-kl]=Vector3(I[j1-1][k].e1(),I[j1][k].e2(),I[j1][k].e3());
    }

// Real region right
    else {
      lBsend[k-kl]=Vector3(0,intBdS[j1][k].e2(),intBdS[j1][k].e3());
      lEsend[k-kl]=Vector3(intEdl[j1][k].e1(), intEdl[j1+1][k].e2(),
		intEdl[j1+1][k].e3());
      lJsend[k-kl]=I[j1][k];
    }

  }

/*
#ifdef MPI_VERSION
  cout << "putCharge_and_Current sending fields:" << endl;
  cout << "E1 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lEsend[kk].e1();
  cout << endl;
  cout << "E2 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lEsend[kk].e2();
  cout << endl;
  cout << "E3 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lEsend[kk].e3();
  cout << endl;
  cout << "B1 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lBsend[kk].e1();
  cout << endl;
  cout << "B2 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lBsend[kk].e2();
  cout << endl;
  cout << "B3 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lBsend[kk].e3();
  cout << endl;
#endif
*/

// Send the fields through the link
  theLink->sendFields(lEsend,lBsend,lJsend);

}                  

// This is the function which does the fieldsolve.
// It does it in such a way that toNodes will be wrong,
// and it does it in such a way that serial operation is
// possible.
// By convention for this boundary, normal points IN

void SpatialRegionBoundary::applyFields(Scalar t,Scalar dt) {

  Scalar dt2 = dt/2;
  Vector3 **intBdS = fields->getIntBdS();
  Vector3 **intEdl = fields->getIntEdl();
  Vector3 **I = fields->getI();
  Vector3 **iC = fields->getiC();
  Vector3 **iL = fields->getiL();

  int kl = MIN(k1,k2);
  int kh = MAX(k1,k2);
  int jl = MIN(j1,j2);
  int jh = MAX(j1,j2);
  int K = fields->getK();
  int J = fields->getJ();
  if(!theLink) return;
  theLink->waitFields();  //  do a synchronize

/*
#ifdef MPI_VERSION
  cout << "applyFields received fields:" << endl;
  cout << "E1 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lE[kk].e1();
  cout << endl;
  cout << "E2 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lE[kk].e2();
  cout << endl;
  cout << "E3 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lE[kk].e3();
  cout << endl;
  cout << "B1 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lB[kk].e1();
  cout << endl;
  cout << "B2 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lB[kk].e2();
  cout << endl;
  cout << "B3 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lB[kk].e3();
  cout << endl;
#endif
*/

  //  Advance B 1/2 step
  if(k1==k2) {  //horizontal boundary 
	 for(int j=jl;j<jh;j++) {
		//The regular field solve handles B2[j][k]
		if(normal == 1) { //  the virtual region is below us
		  lB[j-jl] -= dt2 * Vector3 ( intEdl[j][k1].e3() - lE[j-jl].e3(),
											  0,  //handled by regular field solve
											  lE[j-jl].e1() - intEdl[j][k1].e1() +
											  lE[j-jl+1].e2() - lE[j-jl].e2());
		}
		else {    // the virtual region is above us

		  lB[j-jl] -= dt2 * Vector3 ( -intEdl[j][k1].e3() + lE[j-jl].e3(),
											  0,
											  -lE[j-jl].e1() + intEdl[j][k1].e1() +
											  lE[j-jl+1].e2() - lE[j-jl].e2());
		}
	 }
  }
  else {  //vertical boundary
	 for(int k=kl;k<kh;k++) {
		//The regular field solve handles B1[j][k]
		if(normal == 1) {  // lB handles the virtual fields on the left
		  lB[k-kl] -= dt2 * Vector3(0,  // first component handled by reg. field solve
								- intEdl[j1][k].e3() + lE[k-kl].e3(),
								+ lE[k-kl].e1() - lE[k-kl+1].e1()
								+ intEdl[j1][k].e2() - lE[k-k1].e2());
		}
		else {  //lB handles the virutal fields on the right
		  lB[k-kl] -= dt2 * Vector3(0,  // first component handled by reg. field solve
								- lE[k-kl].e3() + intEdl[j1][k].e3(),
								+ lE[k-kl].e1() - lE[k-kl+1].e1()
								+ lE[k-kl].e2() - intEdl[j1][k].e2());
		}
		  
		// no update of intBds necessary, lB's are all we need, lB1 done properly by reg.
		// field solve.
	 }
  }

  // Advance E 1 step

  if(k1==k2) {  //horizontal boundary 
	 for(int j=jl;j<jh;j++) {
		if(normal==1) {  //real region is below
		  //first update the current
		  I[j][k1]+=Vector3(lJ[j-jl].e1(),0,lJ[j-jl].e3());

		  //update the E fields
		  intEdl[j][k1]+=dt * iC[j][k1-1].jvMult(Vector3(
						((j<J)?lB[j-jl].e3() * liL[j-jl].e3() - intBdS[j][k1-1].e3() * iL[j][k1-1].e3():0),
						0,  //This field isn't my problem here, fields takes care of it
						+intBdS[j][k1-1].e1() * iL[j][k1-1].e1() - lB[j-jl].e1() * liL[j-jl].e1()
						+((j<J)?intBdS[j][k1].e2() * iL[j][k1].e2():0) - ((j>0)?intBdS[j-1][k1].e2()*iL[j-1][k1].e2():0))
						 -Vector3(I[j][k1].e1(),0,I[j][k1].e3()));

		  // need to update this component for to-nodes.
//		  lE[k-kl]+= dt * iC[j1+1][k].jvMult(Vector3(lB[k-kl].e3()*iL[j1+1][k].e3()
//						-((k>0)?lB[k-kl-1].e3()*iL[j1+1][k-1].e3():0) - lJ[k-kl].e1(),0,0));
		
		}
		else {  // real region is above.
		  //first update the current
		  I[j][k1]+=Vector3(lJ[j-jl].e1(),0,lJ[j-jl].e3());

		  //update the E fields
		  intEdl[j][k1]+=dt * iC[j][k1-1].jvMult(Vector3(
						-((j<J)?lB[j-jl].e3() * liL[j-jl].e3() + intBdS[j][k1].e3() * iL[j][k1].e3():0),
						0,  //This field isn't my problem here, fields takes care of it
						-intBdS[j][k1-1].e1() * iL[j][k1-1].e1() + lB[j-jl].e1() * liL[j-jl].e1()
						+((j<J)?intBdS[j][k1].e2() * iL[j][k1].e2():0) - ((j>0)?intBdS[j-1][k1].e2()*iL[j-1][k1].e2():0))
						 -Vector3(I[j][k1].e1(),0,I[j][k1].e3()));

		  
		  
		}
	 }
	  
  }
  else {  //vertical boundary
	 for(int k=kl;k<=kh;k++) {
		if(normal==1) { //real region is to the right.
		  
		  // first update the current 
		  I[j1][k]+=Vector3(0,lJ[k-kl].e2(),lJ[k-kl].e3());
		  
		  intEdl[j1][k]+=dt * iC[j1+1][k].jvMult(Vector3( 0,  //This field isn't my problem here, fields takes care of it
						+((k<K)?-intBdS[j1][k].e3() * iL[j1+1][k].e3() + lB[k-kl].e3()*liL[k-kl].e3():0),
						+((k>0)?intBdS[j1][k-1].e1()* iL[j1+1][k-1].e1():0) - ((k<K)?intBdS[j1][k].e1()*iL[j1+1][k].e1():0) //BdS, iL dont' match, fix later
						+intBdS[j1][k].e2() * iL[j1+1][k].e2() - lB[k-kl].e2()*liL[k-k1].e2())
						 -Vector3(0,I[j1][k].e2(),I[j1][k].e3()));

		  // need to update this component for to-nodes.
//		  lE[k-kl]+= dt * iC[j1+1][k].jvMult(Vector3(lB[k-kl].e3()*iL[j1+1][k].e3()
//						-((k>0)?lB[k-kl-1].e3()*iL[j1+1][k-1].e3():0) - lJ[k-kl].e1(),0,0));
		}
		else {  // real region to the left.
		  // first update the current
		  I[j1][k]+=Vector3(0,lJ[k-kl].e2(),lJ[k-kl].e3());

		  intEdl[j1][k]+=dt * iC[j1-1][k].jvMult(Vector3( 0, //handled by Fields
						+((k<K)?-lB[k-kl].e3()*liL[k-kl].e3()+intBdS[j1-1][k].e3() * iL[j1-1][k].e3()  :0),
						+((k>0)?intBdS[j1][k-1].e1()* iL[j1-1][k-1].e1():0) - ((k<K)?intBdS[j1][k].e1()*iL[j1-1][k].e1():0)  //BdS, iL dont' match, fix later
						+lB[k-kl].e2() * liL[k-kl].e2() - intBdS[j1-1][k].e2()*iL[j1-1][k].e2())
						-Vector3(0,I[j1][k].e2(),I[j1][k].e3()));

		  // need to update this component for to-nodes.
		  //lE[k-kl]+= dt * iC[j1-1][k].jvMult(Vector3(lB[k-kl].e3()*iL[j1-1][k].e3()
//						-((k>0)?lB[k-kl-1].e3()*iL[j1-1][k-1].e3():0) - lJ[k-kl].e1(),0,0));

		}


	 }
  }

	 //  Advance B 1/2 step  --  this advance UNECESSARY because
	 //  the regular field solve will do it for us, based on
	 //  updated intEdl from above.


}


//  This function sends the iL's the paired boundary will need to
//  complete the field solve.
void SpatialRegionBoundary::send_liL() {
  int j,k;
  Vector3 **iL = fields->getiL();
  if(k1==k2) {
	 for(j=j1;j<j2;j++) {
		if(normal==1)  //send the iL's from above
		  lBsend[j-j1] = iL[j][k1+1];
		else //send the iL's from below
		  lBsend[j-j1] = iL[j][k1-1];
	 }
  }
  else {
	 for(k=k1;k<k2;k++) {
		if(normal==1)  //send the iL's from right
		  lBsend[k-k1] = iL[j1+1][k];
		else //send the iL's from left
		  lBsend[k-k1] = iL[j1-1][k];
	 }
  }
  //  Actually send the data over.
  theLink->askFields();  //initialize the transfer
  theLink->sendFields(lEsend,lBsend,lJsend); // do our send
  theLink->waitFields();  //synchronize
}

//  This function initializes the passives across the SRB's.
void SpatialRegionBoundary::initialize_passives() {
  int j,k;
  Vector3 **iL = fields->getiL();
  if(k1==k2) {
	 for(j=j1;j<j2;j++) {
		if(normal==1)  //send the iL's from below
		  lBsend[j-j1] = iL[j][k1+1];
		else //send the iL's from above
		  lBsend[j-j1] = iL[j][k1-1];
	 }
  }
  else {
	 for(k=k1;k<k2;k++) {
		if(normal==1)  //send the iL's from right
		  lBsend[k-k1] = iL[j1+1][k];
		else //send the iL's from left
		  lBsend[k-k1] = iL[j1-1][k];
	 }
  }
  //  Actually send the data over.
  theLink->askFields();  //initialize the transfer
  theLink->sendFields(lEsend,lBsend,lJsend); // do our send
  theLink->waitFields();  //synchronize
  get_liL();
}

//  This function sends the iL's the paired boundary will need to
//  complete the field solve.
void SpatialRegionBoundary::get_liL() {
  int j,k;
    if(k1==k2) {
	 for(j=j1;j<j2;j++) {
		if(normal==1) {
		  liL[j-j1] = lB[j-j1];
		  lB[j-j1]=0;
		}
		else {
		  liL[j-j1] = lB[j-j1];
		  lB[j-j1]=0;
		}
	 }
  }
  else {
	 for(k=k1;k<k2+1;k++) {
		if(normal==1) { 
		  liL[k-k1] = lB[k-k1];
		  lB[k-k1]=0;
		}
		else {
		  liL[k-k1] = lB[k-k1];
		  lB[k-k1]=0;
		}
	 }
  }
}
  

SpatialRegionBoundary::~SpatialRegionBoundary() {
  delete [] lJ;
  delete [] lE;
  delete [] lB;
  lJ=lE=lB=0;
}

//
//  This is the function which packages and sends particles over 
// to the paired SRB.
//
	 
void SpatialRegionBoundary::passParticles() {
  theLink->sendParticles(number_Particles_passing,ptclPassArray);
}

void SpatialRegionBoundary::collect(Particle &p, Vector3& dxMKS) {

// check and see if we've exceeded the passing array memory
  while(number_Particles_passing+1 > ptclPassArraySize) {
    particlePassDat *array = new particlePassDat[2*ptclPassArraySize];
    memcpy(array, ptclPassArray,
	number_Particles_passing * sizeof(particlePassDat));
    delete[] ptclPassArray;
    ptclPassArray = array;
    ptclPassArraySize*=2;
  }

// Add this particle to the array for passing
  ptclPassArray[number_Particles_passing].u = p.get_u();
  if ( p.isVariableWeight() )
       ptclPassArray[number_Particles_passing].vary_np2c = 1.;
  else ptclPassArray[number_Particles_passing].vary_np2c = 0.;
  ptclPassArray[number_Particles_passing].np2c = p.get_np2c();
  ptclPassArray[number_Particles_passing].speciesID = (Scalar) p.get_speciesID();
  ptclPassArray[number_Particles_passing].dxMKS = dxMKS;
  ptclPassArray[number_Particles_passing].x =
     (j1==j2) ? p.get_x().e2() : p.get_x().e1();

// Diagnostics
/*
#ifdef MPI_VERSION
  if ( number_Particles_passing == 0) {
    extern int MPI_RANK;
    cout << "MPI rank " << MPI_RANK << " collected a particle with\n";
    cout << "\n";
  }
#endif
*/

  delete &p;
  number_Particles_passing++;

}

// Emit the particles coming across the boundary.  In this case
// the particles come in and must be weighted in the incoming region.
// This method assumes that calling routine/class will take
// care of emptying the returned ParticleList.  Thus, multiple
// calls to recvParticles accumulate particles.

ParticleList &SpatialRegionBoundary::emit(Scalar t, Scalar dt, 
	Species *species) {

  int i,n_in;
  particlePassDat *inParticles = theLink->waitParticles(&n_in);
  Boundary *bPtr;

  number_Particles_passing=0;  // our send has completed here.
#ifdef DEBUG
  printf("\nSRB:emit:  emitting %d particles",n_in);
#endif

  for(i=0;i<n_in;i++) {

// Get the particle's coordinate
    Vector2 x( (j1==j2)? j1* (1 +normal * 1e-6) +
	normal * 1e-20:inParticles[i].x, 
	(j1==j2) ? inParticles[i].x : k1*(1+normal*1e-6)+normal * 1e-20);

	 // figure out if it is a variable-weight particle
	 BOOL _vary_np2c = FALSE;
	 if(inParticles[i].vary_np2c > 0.) _vary_np2c = TRUE;


    Particle * p = new Particle(x,inParticles[i].u, 
	speciesArray[(int)(inParticles[i].speciesID+0.1)],
	inParticles[i].np2c, _vary_np2c);

// finish the push
    x = p->get_x();
    if((bPtr = fields->translateAccumulate(x, inParticles[i].dxMKS, 
		p->get_q()/fields->get_dt()))) {
      p->set_x(x);
// Send this particle to boundary
      bPtr->collect(*p,inParticles[i].dxMKS);

    }
    else {
      p->set_x(x);
      particleList.add(p);
    }

  }

  return particleList;
}

/**
 * Collect the particle into the shiftedParticles array.
 * This is a handoff, SpatialRegionBoundary is subsequently
 * responsible for deletion of the particle.
 *
 * @param p the particle to be collected
 */
void SpatialRegionBoundary::collectShiftedParticle(Particle* p){
  shiftedParticles.push_back(p);
}

/**
 * Pass the shifted particles over the boundary
 *
 * Because we are responsible for deleting the particles,
 * we must delete after sending.
 */
void SpatialRegionBoundary::passShiftedParticles(){

// Send the particles to the link
/*
#ifdef MPI_VERSION
  extern int MPI_RANK;
  cerr << "MPI_RANK of " << MPI_RANK << ": sending " << 
	shiftedParticles.size() << " particles.\n";
#endif
*/

  if(theLink) theLink->sendShiftedParticles(shiftedParticles);

// Clear out the particles
  for(int i=0; i<shiftedParticles.size(); i++) delete shiftedParticles[i];
  shiftedParticles.clear();

}

/**
 * Receive the shifted particles from the boundary
 */
ParticleList& SpatialRegionBoundary::recvShiftedParticles(){

#ifdef MPI_VERSION
  extern int MPI_RANK;
#endif

// Must have a link
  assert(theLink);

// Return the particles from the link
  // cout << "MPI_RANK = " << MPI_RANK << 
	// ": receiving particles in recvShiftedParticles." << endl;
  ParticleList& pl = theLink->recvShiftedParticles(speciesArray);
  // cout << "MPI_RANK = " << MPI_RANK << 
	// ": particles received in recvShiftedParticles." << endl;

/*
#ifdef MPI_VERSION
  extern int MPI_RANK;
  cerr << "MPI_RANK of " << MPI_RANK << ": received " << 
	pl.nItems() << " particles.\n";
#endif
*/

// Shift the particles over to the right
  oopicListIter<Particle> piter(pl);
  Particle* pp;
  Vector2 x;
  bool prtd = false;
  for(piter.restart(); !piter.Done(); piter++){
    pp = piter.current();
    // if(!prtd){
      // cout << "Before shift, first particle received has x1 = " << 
	// pp->get_x().e1() << endl;
    // }
    x = pp->get_x();
    x.set_e1(x.e1() + j2);
    pp->set_x(x);

    // if(!prtd){
      // cout << "After shift, first particle received has x1 = " << 
	// pp->get_x().e1() << endl;
    // }
    prtd = true;
  }

  return pl;

}

// Ask the link to get the stripe of fields
//
void SpatialRegionBoundary::askStripe(){
  assert(theLink);
  theLink->askShiftedFields();
}


// This is like putCharge_and_Current, except that it sends over
// a pure stripe as is needed for shifting fields
//
void SpatialRegionBoundary::sendStripe(int dir){

#ifdef  MPI_VERSION
  extern int MPI_RANK;
#endif

  Vector3 **intBdS = fields->getIntBdS();
  Vector3 **intEdl = fields->getIntEdl();
  Vector3 **I = fields->getI();

  switch(dir){

// Fields sending to left (right moving window)
    case 1:{
      assert(j1 == j2);
// On KC's authority, one always has k1 <= k2
      for(int k=k1; k<=k2; k++) {
	lEsend[k-k1] = Vector3( intEdl[j1][k].e1(), intEdl[j1+1][k].e2(),
		intEdl[j1+1][k].e3() );
	lBsend[k-k1] = Vector3( intBdS[j1+1][k].e1(), intBdS[j1][k].e2(),
		intBdS[j1][k].e3() );
	lJsend[k-k1] = Vector3( I[j1][k].e1(), I[j1+1][k].e2(),
		I[j1+1][k].e3() );
      }
    }
    break;

// Fields sending to right (left moving window)
    case 2:{
      assert(j1 == j2);
// On KC's authority, one always has k1 <= k2
      for(int k=k1; k<=k2; k++) {
        lEsend[k-k1] = intEdl[j1-1][k];
        lBsend[k-k1] = intBdS[j1-1][k];
        lJsend[k-k1] = I[j1-1][k];
      }
    }
    break;

// Fields sending up (to smaller k, downward moving window)
    case 3:{
      assert(k1 == k2);
// On KC's authority, one always has k1 <= k2
      for(int j=j1; j<=j2; j++) {
	lEsend[j-j1] = Vector3( intEdl[j][k1+1].e1(), intEdl[j][k1].e2(),
		intEdl[j][k1+1].e3() );
	lBsend[j-j1] = Vector3( intBdS[j][k1].e1(), intBdS[j][k1+1].e2(),
		intBdS[j][k1].e3() );
	lJsend[j-j1] = Vector3( I[j][k1+1].e1(), I[j][k1].e2(),
		I[j][k1+1].e3() );
      }
    }
    break;

// Fields sending down (to greater k, upward moving window)
    case 4:{
      assert(k1 == k2);
// On KC's authority, one always has k1 <= k2
      for(int j=j1; j<=j2; j++) {
        lEsend[j-j1] = intEdl[j][k1-1];
        lBsend[j-j1] = intBdS[j][k1-1];
        lJsend[j-j1] = I[j][k1-1];
      }
    }
    break;

  }

/*
#ifdef MPI_VERSION
  cout << "Sending stripe:" << endl;
  cout << "E1 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lEsend[kk].e1();
  cout << endl;
  cout << "E2 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lEsend[kk].e2();
  cout << endl;
  cout << "E3 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lEsend[kk].e3();
  cout << endl;
  cout << "B1 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lBsend[kk].e1();
  cout << endl;
  cout << "B2 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lBsend[kk].e2();
  cout << endl;
  cout << "B3 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lBsend[kk].e3();
  cout << endl;
#endif
*/

// Send the fields through the link
/*
#ifdef  MPI_VERSION
  cout << "MPI_RANK = " << MPI_RANK << 
	", SpatialRegionBoundary::sendStripe sending fields. j1,j2,k1,k2 = " <<
	j1 << "," << j2 << "," << k1 << "," << k2 << endl;
#endif
*/

  theLink->sendShiftedFields(lEsend,lBsend,lJsend);

/*
#ifdef  MPI_VERSION
  cout << "MPI_RANK = " << MPI_RANK << 
	", SpatialRegionBoundary::sendStripe sent fields." << endl;
#endif
*/

}

void SpatialRegionBoundary::recvStripe(int dir){

  Vector3 **intBdS = fields->getIntBdS();
  Vector3 **intEdl = fields->getIntEdl();
  Vector3 **I = fields->getI();

  assert(theLink);
#ifdef  MPI_VERSION
  extern int MPI_RANK;
#endif

/*
#ifdef  MPI_VERSION
  cout << "MPI_RANK = " << MPI_RANK << 
	", SpatialRegionBoundary::recvStripe awaiting fields. j1,j2,k1,k2 = " <<
	j1 << "," << j2 << "," << k1 << "," << k2 << endl;
#endif
*/

  theLink->waitShiftedFields();  //  do a synchronize

/*
#ifdef  MPI_VERSION
  cout << "MPI_RANK = " << MPI_RANK << 
	", SpatialRegionBoundary::recvStripe received fields." << endl;
#endif
*/

/*
#ifdef MPI_VERSION
  cout << "Received stripe:" << endl;
  cout << "E1 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lE[kk].e1();
  cout << endl;
  cout << "E2 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lE[kk].e2();
  cout << endl;
  cout << "E3 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lE[kk].e3();
  cout << endl;
  cout << "B1 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lB[kk].e1();
  cout << endl;
  cout << "B2 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lB[kk].e2();
  cout << endl;
  cout << "B3 =";
  for(int kk=0; kk<=k2-k1; kk++) cout << " " << lB[kk].e3();
  cout << endl;
#endif
*/

  switch(dir){

// Fields coming from right (right moving window)
    case 1:{
      assert(j1 == j2);
// On KC's authority, one always has k1 <= k2
      for(int k=k1; k<=k2; k++) {
        intEdl[j1-1][k].set_e1(lE[k-k1].e1());
        intEdl[j1][k].set_e2(lE[k-k1].e2());
        intEdl[j1][k].set_e3(lE[k-k1].e3());
        intBdS[j1][k].set_e1(lB[k-k1].e1());
        intBdS[j1-1][k].set_e2(lB[k-k1].e2());
        intBdS[j1-1][k].set_e3(lB[k-k1].e3());
        I[j1-1][k].set_e1(lJ[k-k1].e1());
        I[j1][k].set_e2(lJ[k-k1].e2());
        I[j1][k].set_e3(lJ[k-k1].e3());
      }
    }
    break;

// Fields coming from left (left moving window)
    case 2:{
      assert(j1 == j2);
// On KC's authority, one always has k1 <= k2
      for(int k=k1; k<=k2; k++) {
        intEdl[j1][k] = lE[k-k1];
        intBdS[j1][k] = lB[k-k1];
        I[j1][k] = lJ[k-k1];
      }
    }
    break;

// Fields coming from bottom (from greater k, downward moving window)
    case 3:{
      assert(k1 == k2);
// On KC's authority, one always has k1 <= k2
      for(int j=j1; j<=j2; j++) {
        intEdl[j][k1].set_e1(lE[j-j1].e1());
        intEdl[j][k1-1].set_e2(lE[j-j1].e2());
        intEdl[j][k1].set_e3(lE[j-j1].e3());
        intBdS[j][k1-1].set_e1(lB[j-j1].e1());
        intBdS[j][k1].set_e2(lB[j-j1].e2());
        intBdS[j][k1-1].set_e3(lB[j-j1].e3());
        I[j][k1].set_e1(lJ[j-j1].e1());
        I[j][k1-1].set_e2(lJ[j-j1].e2());
        I[j][k1].set_e3(lJ[j-j1].e3());
      }
    }
    break;

// Fields coming from top (from smaller k, upward moving window)
    case 4:{
      assert(k1 == k2);
// On KC's authority, one always has k1 <= k2
      for(int j=j1; j<=j2; j++) {
        intEdl[j][k1] = lE[j-j1];
        intBdS[j][k1] = lB[j-j1];
        I[j][k1] = lJ[j-j1];
      }
    }
    break;

  }

}
